# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr
import ctypes as ct

WSADESCRIPTION_LEN = 256
WSASYS_STATUS_LEN = 128

MAX_PATH = 260
MAX_MODULE_NAME32 = 255

FILE_ATTRIBUTE_NORMAL = 0x80

TH32CS_INHERIT = 0x80000000
TH32CS_SNAPHEAPLIST = 0x00000001
TH32CS_SNAPMODULE = 0x00000008
TH32CS_SNAPMODULE32 = 0x00000010
TH32CS_SNAPPROCESS = 0x00000002
TH32CS_SNAPTHREAD = 0x00000004

PROCESSOR_ARCHITECTURE_AMD64 = 9
PROCESSOR_ARCHITECTURE_INTEL = 0

LOCALE_INVARIANT = 0x7F
LOCALE_USER_DEFAULT = 0x400
LOCALE_SYSTEM_DEFAULT = 0x800
LOCALE_CUSTOM_DEFAULT = 0xC00
LOCALE_CUSTOM_UNSPECIFIED = 0x1000
LOCALE_CUSTOM_UI_DEFAULT = 0x1400

LOCALE_SENGLISHLANGUAGENAME = 0x1001
LOCALE_SENGLISHCOUNTRYNAME = 0x1002

DRIVE_UNKNOWN = 0
DRIVE_NO_ROOT_DIR = 1
DRIVE_REMOVABLE = 2
DRIVE_FIXED = 3
DRIVE_REMOTE = 4
DRIVE_CDROM = 5
DRIVE_RAMDISK = 6

ComputerNameNetBIOS = 0
ComputerNameDnsHostname = 1
ComputerNameDnsDomain = 2
ComputerNameDnsFullyQualified = 3
ComputerNamePhysicalNetBIOS = 4
ComputerNamePhysicalDnsHostname = 5
ComputerNamePhysicalDnsDomain = 6
ComputerNamePhysicalDnsFullyQualified = 7
ComputerNameMax = 8

GetFileExInfoStandard = 0

EXCEPTION_CONTINUE_SEARCH = 0
EXCEPTION_EXECUTE_HANDLER = 1

THREAD_PRIORITY_NORMAL = 0

class PROCESSENTRY32(EmuStruct):
    def __init__(self, ptr_size, width):
        super().__init__(ptr_size)
        self.dwSize = ct.c_uint32
        self.cntUsage = ct.c_uint32
        self.th32ProcessID = ct.c_uint32
        self.th32DefaultHeapID = Ptr
        self.th32ModuleID = ct.c_uint32
        self.cntThreads = ct.c_uint32
        self.th32ParentProcessID = ct.c_uint32
        self.pcPriClassBase = ct.c_uint32
        self.dwFlags = ct.c_uint32
        self.szExeFile = ct.c_uint8 * (MAX_PATH * width)


class THREADENTRY32(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwSize = ct.c_uint32
        self.cntUsage = ct.c_uint32
        self.th32ThreadID = ct.c_uint32
        self.th32OwnerProcessID = ct.c_uint32
        self.tpBasePri = ct.c_uint32
        self.tpDeltaPri = ct.c_uint32
        self.dwFlags = ct.c_uint32


class MODULEENTRY32(EmuStruct):
    def __init__(self, ptr_size, width):
        super().__init__(ptr_size)
        self.dwSize = ct.c_uint32
        self.th32ModuleID = ct.c_uint32
        self.th32ProcessID = ct.c_uint32
        self.GlblcntUsage = ct.c_uint32
        self.ProccntUsage = ct.c_uint32
        self.modBaseAddr = Ptr
        self.modBaseSize = ct.c_uint32
        self.hModule = ct.c_uint32
        self.szModule = ct.c_uint8 * ((MAX_MODULE_NAME32 + 1) * width)
        self.szExePath = ct.c_uint8 * (MAX_PATH * width)


class PROCESS_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.hProcess = Ptr
        self.hThread = Ptr
        self.dwProcessId = ct.c_uint32
        self.dwThreadId = ct.c_uint32


class MEMORY_BASIC_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.BaseAddress = Ptr
        self.AllocationBase = Ptr
        self.AllocationProtect = ct.c_uint32
        self.RegionSize = Ptr
        self.State = ct.c_uint32
        self.Protect = ct.c_uint32
        self.Type = ct.c_uint32


class FILETIME(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwLowDateTime = ct.c_uint32
        self.dwHighDateTime = ct.c_uint32


class WIN32_FIND_DATA(EmuStruct):
    def __init__(self, ptr_size, width):
        super().__init__(ptr_size)
        self.dwFileAttributes = ct.c_uint32
        self.ftCreationTime = FILETIME
        self.ftLastAccessTime = FILETIME
        self.ftLastWriteTime = FILETIME
        self.nFileSizeHigh = ct.c_uint32
        self.nFileSizeLow = ct.c_uint32
        self.dwReserved0 = ct.c_uint32
        self.dwReserved1 = ct.c_uint32
        self.cFileName = ct.c_uint8 * (260 * width)
        self.cAlternateFileName = ct.c_uint8 * (14 * width)


class WIN32_FILE_ATTRIBUTE_DATA(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwFileAttributes = ct.c_uint32
        self.ftCreationTime = FILETIME
        self.ftLastAccessTime = FILETIME
        self.ftLastWriteTime = FILETIME
        self.nFileSizeHigh = ct.c_uint32
        self.nFileSizeLow = ct.c_uint32


class SYSTEM_INFO(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.wProcessorArchitecture = ct.c_uint16
        self.dwPageSize = ct.c_uint32
        self.lpMinimumApplicationAddress = Ptr
        self.lpMaximumApplicationAddress = Ptr
        self.dwActiveProcessorMask = Ptr
        self.dwNumberOfProcessors = ct.c_uint32
        self.dwProcessorType = ct.c_uint32
        self.dwAllocationGranularity = ct.c_uint32
        self.wProcessorLevel = ct.c_uint16
        self.wProcessorRevision = ct.c_uint16


class SYSTEMTIME(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.wYear = ct.c_uint16
        self.wMonth = ct.c_uint16
        self.wDayOfWeek = ct.c_uint16
        self.wDay = ct.c_uint16
        self.wHour = ct.c_uint16
        self.wMinute = ct.c_uint16
        self.wSecond = ct.c_uint16
        self.wMilliseconds = ct.c_uint16


class STARTUPINFO(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.cb = ct.c_uint32
        self.lpReserved = Ptr
        self.lpDesktop = Ptr
        self.lpTitle = Ptr
        self.dwX = ct.c_uint32
        self.dwY = ct.c_uint32
        self.dwXSize = ct.c_uint32
        self.dwYSize = ct.c_uint32
        self.dwXCountChars = ct.c_uint32
        self.dwYCountChars = ct.c_uint32
        self.dwFillAttribute = ct.c_uint32
        self.dwFlags = ct.c_uint32
        self.wShowWindow = ct.c_uint16
        self.cbReserved2 = ct.c_uint16
        self.lpReserved2 = Ptr
        self.hStdInput = Ptr
        self.hStdOutput = Ptr
        self.hStdError = Ptr


class OSVERSIONINFO(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwOSVersionInfoSize = ct.c_uint32
        self.dwMajorVersion = ct.c_uint32
        self.dwMinorVersion = ct.c_uint32
        self.dwBuildNumber = ct.c_uint32
        self.dwPlatformId = ct.c_uint32
        self.szCSDVersion = ct.c_uint8 * 128


class OSVERSIONINFOEX(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwOSVersion = OSVERSIONINFO
        self.dwPlatformId = ct.c_uint16
        self.dwPlatformId = ct.c_uint16
        self.wProductType = ct.c_uint8
        self.wReserved = ct.c_uint8


def get_define(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k


def get_define_value(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or k != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return v
        else:
            return v


def get_flag_defines(flags, prefix=''):
    defs = []
    for k, v in globals().items():
        if not isinstance(v, int):
            continue
        if v & flags:
            if prefix and k.startswith(prefix):
                defs.append(k)
    return defs



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import unittest

import util


class TestDllEmulation(unittest.TestCase):

    def setUp(self):
        self.config = util.get_config()
        self.report = None

    def _get_api_calls(self, ep, api_name):
        return [api for api in ep['apis'] if api['api_name'] == api_name]

    def _test_dll_emu(self, fpath):
        data = util.get_test_bin_data(fpath)
        self.report = util.run_test(self.config, data)
        eps = self.report['entry_points']
        self.assertEqual(len(eps), 3)

        dll_entry = eps[0]

        msgbox = self._get_api_calls(dll_entry, 'USER32.MessageBoxA')
        self.assertEqual(1, len(msgbox))
        msgbox = msgbox[0]
        self.assertEqual(msgbox['args'][1], 'Inside process attach')
        self.assertEqual(msgbox['args'][2], 'My caption')
        self.assertEqual(dll_entry['ret_val'], '0x1')

        ep = eps[1]
        msgbox = self._get_api_calls(ep, 'USER32.MessageBoxA')
        self.assertEqual(1, len(msgbox))
        msgbox = msgbox[0]
        self.assertEqual(msgbox['args'][1], 'Inside emu_test_one')
        self.assertEqual(msgbox['args'][2], 'First export')
        self.assertEqual(ep['ret_val'], '0x41414141')

        ep = eps[2]
        msgbox = self._get_api_calls(ep, 'USER32.MessageBoxW')
        self.assertEqual(1, len(msgbox))
        msgbox = msgbox[0]
        self.assertEqual(msgbox['args'][1], 'Inside emu_test_two')
        self.assertEqual(msgbox['args'][2], 'Second export')
        self.assertEqual(ep['ret_val'], '0x42424242')

    def test_x86_dll(self):
        self._test_dll_emu('dll_test_x86.dll.xz')

    def test_x64_dll(self):
        self._test_dll_emu('dll_test_x64.dll.xz')


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch
import speakeasy.windows.sessman as sessman
import speakeasy.winenv.defs.windows.user32 as windefs
import speakeasy.winenv.defs.windows.windef as windef

from .. import api

IDCANCEL = 2

IDI_APPLICATION = 32512
IDI_ASTERISK = 32516
IDI_ERROR = 32513
IDI_EXCLAMATION = 32515
IDI_HAND = 32513
IDI_INFORMATION = 32516
IDI_QUESTION = 32514
IDI_SHIELD = 32518
IDI_WARNING = 32515
IDI_WINLOGO = 32517

UOI_FLAGS = 1


class User32(api.ApiHandler):

    """
    Implements exported functions from user32.dll
    """

    name = 'user32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(User32, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.window_hooks = {}
        self.handle = 0
        self.win = None
        self.handles = []
        self.timer_count = 0
        self.sessman = sessman.SessionManager(config=None)

        super(User32, self).__get_hook_attrs__(self)

    def get_handle(self):
        self.handle += 4
        hnd = self.handle
        self.handles.append(hnd)
        return hnd

    def find_string_resource_by_id(self, pe, uID):
        for entry in pe.DIRECTORY_ENTRY_RESOURCE.entries:
            if entry.id == 6:  # "String Table"
                for str_entry in entry.directory.entries:
                    s = str_entry.directory.strings.get(uID, None)
                    if s is not None:
                        return s
        return None

    @apihook('GetDesktopWindow', argc=0)
    def GetDesktopWindow(self, emu, argv, ctx={}):
        '''HWND GetDesktopWindow();'''

        hnd = 0

        desk = self.sessman.get_current_desktop()
        window = desk.get_desktop_window()
        hnd = window.get_handle()

        return hnd

    @apihook('ShowWindow', argc=2)
    def ShowWindow(self, emu, argv, ctx={}):
        '''BOOL ShowWindow(
          HWND hWnd,
          int  nCmdShow
        );'''

        rv = 1

        return rv

    @apihook('CreateWindowStation', argc=4)
    def CreateWindowStation(self, emu, argv, ctx={}):
        '''
        HWINSTA CreateWindowStation(
            LPCSTR                lpwinsta,
            DWORD                 dwFlags,
            ACCESS_MASK           dwDesiredAccess,
            LPSECURITY_ATTRIBUTES lpsa
        );
        '''
        winsta, flags, access, sa = argv

        return self.get_handle()

    @apihook('SetProcessWindowStation', argc=1)
    def SetProcessWindowStation(self, emu, argv, ctx={}):
        '''
        BOOL SetProcessWindowStation(
            HWINSTA hWinSta
        );
        '''
        winsta, = argv

        rv = False
        if winsta:
            rv = True

        return rv

    @apihook('GetDC', argc=1)
    def GetDC(self, emu, argv, ctx={}):
        '''
        HDC GetDC(
          HWND hWnd
        );
        '''

        rv = self.sessman.get_device_context()

        return rv

    @apihook('RegisterClassEx', argc=1)
    def RegisterClassEx(self, emu, argv, ctx={}):
        '''
        ATOM RegisterClassEx(
            const WNDCLASSEXA *Arg1
        );
        '''
        Arg1, = argv
        wclass = windefs.WNDCLASSEX(emu.get_ptr_size())
        wclass = self.mem_cast(wclass, Arg1)

        cn = None
        cw = self.get_char_width(ctx)
        if wclass.lpszClassName:
            cn = self.read_mem_string(wclass.lpszClassName, cw)

        atom = self.sessman.create_window_class(wclass, cn)

        return atom

    @apihook('UnregisterClass', argc=2)
    def UnregisterClass(self, emu, argv, ctx={}):
        '''
        BOOL UnregisterClass(
            LPCSTR    lpClassName,
            HINSTANCE hInstance
        );
        '''

        return 1

    @apihook('SetCursorPos', argc=2)
    def SetCursorPos(self, emu, argv, ctx={}):
        '''
        BOOL SetCursorPos(
        int X,
        int Y
        );
        '''
        return 1

    @apihook('CloseDesktop', argc=1)
    def CloseDesktop(self, emu, argv, ctx={}):
        '''
        BOOL CloseDesktop(
        HDESK hDesktop
        );
        '''
        return 1

    @apihook('CloseWindowStation', argc=1)
    def CloseWindowStation(self, emu, argv, ctx={}):
        '''
        BOOL CloseWindowStation(
        HWINSTA hWinSta
        );
        '''
        return 1

    @apihook('GetThreadDesktop', argc=1)
    def GetThreadDesktop(self, emu, argv, ctx={}):
        '''
        HDESK GetThreadDesktop(
        DWORD dwThreadId
        );
        '''
        return 1

    @apihook('OpenWindowStation', argc=3)
    def OpenWindowStation(self, emu, argv, ctx={}):
        '''
        HWINSTA OpenWindowStation(
        LPCSTR      lpszWinSta,
        BOOL        fInherit,
        ACCESS_MASK dwDesiredAccess
        );
        '''
        return 1

    @apihook('ChangeWindowMessageFilter', argc=2)
    def ChangeWindowMessageFilter(self, emu, argv, ctx={}):
        '''
        BOOL ChangeWindowMessageFilter(
            UINT  message,
            DWORD dwFlag
        );
        '''
        msg, flag = argv
        emu.enable_code_hook()
        return True

    @apihook('UpdateWindow', argc=1)
    def UpdateWindow(self, emu, argv, ctx={}):
        '''
        BOOL UpdateWindow(
            HWND hWnd
        );
        '''
        hnd, = argv
        window = self.sessman.get_window(hnd)
        if not window:
            return False

        wc = self.sessman.get_window_class(window.class_name)
        if wc.wclass.lpfnWndProc:
            cb_args = (hnd, windefs.WM_PAINT, 0, 0)
            self.setup_callback(wc.wclass.lpfnWndProc, cb_args, caller_argv=argv)

        return True

    @apihook('PostQuitMessage', argc=1)
    def PostQuitMessage(self, emu, argv, ctx={}):
        '''
        void PostQuitMessage(
            int nExitCode
        );
        '''
        return

    @apihook('DestroyWindow', argc=1)
    def DestroyWindow(self, emu, argv, ctx={}):
        '''
        BOOL DestroyWindow(
            HWND hWnd
        );
        '''
        return True

    @apihook('DefWindowProc', argc=4)
    def DefWindowProc(self, emu, argv, ctx={}):
        '''
        LRESULT LRESULT DefWindowProc(
            HWND   hWnd,
            UINT   Msg,
            WPARAM wParam,
            LPARAM lParam
        );
        '''
        return 0

    @apihook('CreateWindowEx', argc=12)
    def CreateWindowEx(self, emu, argv, ctx={}):
        '''
        HWND CreateWindowExA(
            DWORD     dwExStyle,
            LPCSTR    lpClassName,
            LPCSTR    lpWindowName,
            DWORD     dwStyle,
            int       X,
            int       Y,
            int       nWidth,
            int       nHeight,
            HWND      hWndParent,
            HMENU     hMenu,
            HINSTANCE hInstance,
            LPVOID    lpParam
        );
        '''
        cw = self.get_char_width(ctx)
        _, cn, wn, _, x, y, width, height, parent, menu, inst, param = argv
        if cn:
            cn = self.read_mem_string(cn, cw)
            argv[1] = cn
        else:
            cn = None
        if wn:
            wn = self.read_mem_string(wn, cw)
            argv[2] = wn
        else:
            wn = None
        hnd = self.sessman.create_window(wn, cn)
        return hnd

    @apihook('MessageBox', argc=4)
    def MessageBox(self, emu, argv, ctx={}):
        '''int MessageBox(
          HWND    hWnd,
          LPCTSTR lpText,
          LPCTSTR lpCaption,
          UINT    uType
        );'''
        hWnd, lpText, lpCaption, uType = argv

        cw = self.get_char_width(ctx)

        if lpText:
            text = self.read_mem_string(lpText, cw)
            argv[1] = text
        if lpCaption:
            cap = self.read_mem_string(lpCaption, cw)
            argv[2] = cap
        rv = IDCANCEL

        return rv

    @apihook('MessageBoxEx', argc=5)
    def MessageBoxEx(self, emu, argv, ctx={}):
        '''
        int MessageBoxExA(
            HWND   hWnd,
            LPCSTR lpText,
            LPCSTR lpCaption,
            UINT   uType,
            WORD   wLanguageId
        );
        '''
        av = argv[:-1]
        rv = self.MessageBox(emu, av, ctx)
        argv[:4] = av
        return rv

    @apihook('LoadString', argc=4)
    def LoadString(self, emu, argv, ctx={}):
        """
        int LoadStringW(
          HINSTANCE hInstance,
          UINT      uID,
          LPWSTR    lpBuffer,
          int       cchBufferMax
        );
        """

        hInstance, uID, lpBuffer, ccBufferMax = argv
        cw = self.get_char_width(ctx)
        size = 0

        if hInstance == 0:
            pe = emu.modules[0][0]
        else:
            pe = emu.get_mod_from_addr(hInstance)
            if pe and hInstance != pe.get_base():
                return 0

        s = self.find_string_resource_by_id(pe, uID)
        if s is None:
            # self.logger.info("unable to find resource string id %04X" % uID)
            return 0

        if cw == 2:
            encoded = s.encode('utf-16le')
        elif cw == 1:
            encoded = s.encode('utf-8')

        size = int(len(encoded) / cw)

        if size == 0:
            # self.logger.debug("resource id %04X not found" % uID)
            return 0

        if ccBufferMax == 0:
            # TODO this should be done properly, but requires more research
            offset = pe.get_memory_mapped_image().find(encoded)
            emu.mem_write(lpBuffer, (pe.get_base() + offset).to_bytes(emu.get_ptr_size(),
                                                                      'little'))
            return size

        if len(encoded) > ccBufferMax:
            encoded = encoded[:ccBufferMax * cw]

        emu.mem_write(lpBuffer, encoded)
        if cw == 1:
            argv[2] = s
        else:
            argv[2] = s

        return len(encoded)

    @apihook('GetCursorPos', argc=1)
    def GetCursorPos(self, emu, argv, ctx={}):
        """
        BOOL GetCursorPos(
          LPPOINT lpPoint
        );
        """

        lpPoint, = argv

        rv = 0
        return rv

    @apihook('GetKeyboardType', argc=1)
    def GetKeyboardType(self, emu, argv, ctx={}):
        '''
        int GetKeyboardType(
          int nTypeFlag
        );
        '''
        _type, = argv
        if _type == 0:
            return 4
        elif _type == 1:
            return 0
        elif _type == 2:
            return 12
        return 0

    @apihook('GetSystemMetrics', argc=1)
    def GetSystemMetrics(self, emu, argv, ctx={}):
        """
        int GetSystemMetrics(
          int nIndex
        );
        """

        nIndex, = argv

        rv = 1
        return rv

    @apihook('LoadBitmap', argc=2)
    def LoadBitmap(self, emu, argv, ctx={}):
        """
        HBITMAP LoadBitmap(
            HINSTANCE hInstance,
            LPCSTR    lpBitmapName
        );
        """
        hInstance, lpBitmapName = argv
        rv = self.get_handle()
        return rv

    @apihook('RegisterWindowMessage', argc=1)
    def RegisterWindowMessage(self, emu, argv, ctx={}):
        '''
        UINT RegisterWindowMessageA(
          LPCSTR lpString
        );
        '''

        lpString, = argv
        rv = 0xc000

        cw = self.get_char_width(ctx)

        s = self.read_mem_string(lpString, cw)
        argv[0] = s

        return rv

    @apihook('wsprintf', argc=_arch.VAR_ARGS, conv=_arch.CALL_CONV_CDECL)
    def wsprintf(self, emu, argv, ctx={}):
        """
        int WINAPIV wsprintf(
          LPSTR  ,
          LPCSTR ,
          ...
        );
        """
        cw = self.get_char_width(ctx)

        buf, fmt = emu.get_func_argv(_arch.CALL_CONV_CDECL, 2)
        fmt_str = self.read_mem_string(fmt, cw)
        fmt_cnt = self.get_va_arg_count(fmt_str)

        if not fmt_cnt:
            self.write_mem_string(fmt_str, buf, cw)
            return len(fmt_str)

        _args = emu.get_func_argv(_arch.CALL_CONV_CDECL, 2 + fmt_cnt)[2:]
        fin = self.do_str_format(fmt_str, _args)

        self.write_mem_string(fin, buf, cw)

        argv.append(fin)
        argv.append(fmt_str)
        return len(fin)

    @apihook('PeekMessage', argc=5)
    def PeekMessage(self, emu, argv, ctx={}):
        '''
        BOOL PeekMessageA(
            LPMSG lpMsg,
            HWND  hWnd,
            UINT  wMsgFilterMin,
            UINT  wMsgFilterMax,
            UINT  wRemoveMsg
        );
        '''
        return False

    @apihook('PostMessage', argc=4)
    def PostMessage(self, emu, argv, ctx={}):
        '''
        BOOL PostMessage(
            HWND   hWnd,
            UINT   Msg,
            WPARAM wParam,
            LPARAM lParam
        );
        '''
        return True

    @apihook('SendMessage', argc=4)
    def SendMessage(self, emu, argv, ctx={}):
        '''
        LRESULT SendMessage(
            HWND   hWnd,
            UINT   Msg,
            WPARAM wParam,
            LPARAM lParam
        );
        '''
        return False

    @apihook('SetWindowsHookEx', argc=4)
    def SetWindowsHookEx(self, emu, argv, ctx={}):
        '''
        HHOOK SetWindowsHookEx(
            int       idHook,
            HOOKPROC  lpfn,
            HINSTANCE hmod,
            DWORD     dwThreadId
        );
        '''
        idHook, lpfn, hmod, dwThreadId = argv

        hname = windefs.get_windowhook_flags(idHook)
        if hname:
            hname = hname[0]
            argv[0] = hname

        hnd = self.get_handle()
        self.window_hooks.update({hnd: (lpfn, hmod)})
        return hnd

    @apihook('UnhookWindowsHookEx', argc=1)
    def UnhookWindowsHookEx(self, emu, argv, ctx={}):
        '''
        BOOL UnhookWindowsHookEx(
            HHOOK hhk
        );
        '''
        hhk, = argv

        rv = False
        if self.window_hooks.get(hhk):
            self.window_hooks.pop(hhk)
            rv = True
        return rv

    @apihook('MsgWaitForMultipleObjects', argc=5)
    def MsgWaitForMultipleObjects(self, emu, argv, ctx={}):
        '''
        DWORD MsgWaitForMultipleObjects(
            DWORD        nCount,
            const HANDLE *pHandles,
            BOOL         fWaitAll,
            DWORD        dwMilliseconds,
            DWORD        dwWakeMask
        );
        '''
        return 0

    @apihook('GetMessage', argc=4)
    def GetMessage(self, emu, argv, ctx={}):
        '''
        BOOL GetMessage(
            LPMSG lpMsg,
            HWND  hWnd,
            UINT  wMsgFilterMin,
            UINT  wMsgFilterMax
        );
        '''
        lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax = argv

        t = emu.get_current_thread()
        try:
            msg = t.message_queue.pop(0)
        except IndexError:
            # If the queue is empty but a timer is active, write a WM_TIMER message and return True
            if self.timer_count > 0:
                msg = windefs.MSG(emu.get_ptr_size())
                msg.hwnd = hWnd
                msg.message = windefs.WM_TIMER
            else:
                return False

        self.mem_write(lpMsg, msg.get_bytes())

        return True

    @apihook('TranslateMessage', argc=1)
    def TranslateMessage(self, emu, argv, ctx={}):
        '''
        BOOL TranslateMessage(
            const MSG *lpMsg
        );
        '''
        return True

    @apihook('DispatchMessage', argc=1)
    def DispatchMessage(self, emu, argv, ctx={}):
        '''
        LRESULT DispatchMessage(
            const MSG *lpMsg
        );
        '''
        lpMsg, = argv

        msg = windefs.MSG(emu.get_ptr_size())
        msg = self.mem_cast(msg, lpMsg)

        return 0

    @apihook('GetForegroundWindow', argc=0)
    def GetForegroundWindow(self, emu, argv, ctx={}):
        '''
        HWND GetForegroundWindow();
        '''
        return self.get_handle()

    @apihook('LoadCursor', argc=2)
    def LoadCursor(self, emu, argv, ctx={}):
        '''
        HCURSOR LoadCursor(
        HINSTANCE hInstance,
        LPCSTR    lpCursorName
        );
        '''
        return self.get_handle()

    @apihook('FindWindow', argc=2)
    def FindWindow(self, emu, argv, ctx={}):
        '''
        HWND FindWindow(
            LPCSTR lpClassName,
            LPCSTR lpWindowName
        );
        '''
        lpClassName, lpWindowName = argv
        cw = self.get_char_width(ctx)
        if lpClassName:
            cn = self.read_mem_string(lpClassName, cw)
            argv[0] = cn
        if lpWindowName:
            wn = self.read_mem_string(lpWindowName, cw)
            argv[1] = wn
        return 0

    @apihook('GetWindowText', argc=3)
    def GetWindowText(self, emu, argv, ctx={}):
        '''
        int GetWindowText(
            HWND  hWnd,
            LPSTR lpString,
            int   nMaxCount
        );
        '''
        hnd, pstr, maxc = argv

        cw = self.get_char_width(ctx)
        win_text = 'speakeasy window'
        if pstr:
            if cw == 2:
                wt = (win_text).encode('utf-16le')
            else:
                wt = (win_text).encode('utf-8')
            self.mem_write(pstr, wt)

        return len(win_text)

    @apihook('PaintDesktop', argc=1)
    def PaintDesktop(self, emu, argv, ctx={}):
        '''
        BOOL PaintDesktop(
        HDC hdc
        );
        '''
        return 0

    @apihook('wvsprintf', argc=_arch.VAR_ARGS, conv=_arch.CALL_CONV_CDECL)
    def wvsprintf(self, emu, argv, ctx={}):
        buf, fmt, va_list = emu.get_func_argv(_arch.CALL_CONV_CDECL, 3)[:3]
        cw = self.get_char_width(ctx)
        fmt_str = self.read_mem_string(fmt, cw)
        fmt_cnt = self.get_va_arg_count(fmt_str)

        vargs = self.va_args(va_list, fmt_cnt)
        fin = self.do_str_format(fmt_str, vargs)
        self.write_string(fin, buf)
        argv.clear()
        argv.append(fin)
        argv.append(fmt_str)
        return len(fin)

    @apihook('ReleaseDC', argc=2)
    def ReleaseDC(self, emu, argv, ctx={}):
        '''
        int ReleaseDC(
          HWND hWnd,
          HDC  hDC
        );
        '''
        return 0

    @apihook('CharNext', argc=1)
    def CharNext(self, emu, argv, ctx={}):
        '''
        LPSTR CharNext(
            LPCSTR lpsz
        );
        '''
        s, = argv
        rv = 0
        cw = self.get_char_width(ctx)
        if s:
            rv = s + cw
        return rv

    @apihook('EnumWindows', argc=2)
    def EnumWindows(self, emu, argv, ctx={}):
        '''
        BOOL EnumWindows(
            WNDENUMPROC lpEnumFunc,
            LPARAM      lParam
        );
        '''
        lpEnumFunc, lParam = argv
        rv = 1

        return rv

    @apihook('GetSysColor', argc=1)
    def GetSysColor(self, emu, argv, ctx={}):
        '''
        DWORD GetSysColor(
            int nIndex
        );
        '''
        nIndex, = argv
        rv = 1

        return rv

    @apihook('GetParent', argc=1)
    def GetParent(self, emu, argv, ctx={}):
        '''
        HWND GetParent(
            HWND hWnd
        );
        '''
        return self.get_handle()

    @apihook('GetSysColorBrush', argc=1)
    def GetSysColorBrush(self, emu, argv, ctx={}):
        '''
        HBRUSH GetSysColorBrush(
            int nIndex
        );
        '''
        nIndex, = argv
        rv = 1

        return rv

    @apihook('GetWindowLong', argc=2)
    def GetWindowLong(self, emu, argv, ctx={}):
        '''
        LONG GetWindowLongA(
            HWND hWnd,
            int  nIndex
        );
        '''
        hWnd, nIndex, = argv
        rv = 2

        return rv

    @apihook('SetWindowLong', argc=3)
    def SetWindowLong(self, emu, argv, ctx={}):
        """
        LONG SetWindowLongA(
          HWND hWnd,
          int  nIndex,
          LONG dwNewLong
        );
        """


        return 1

    @apihook('DialogBoxParam', argc=5)
    def DialogBoxParam(self, emu, argv, ctx={}):
        '''
        INT_PTR DialogBoxParam(
            HINSTANCE hInstance,
            LPCSTR    lpTemplateName,
            HWND      hWndParent,
            DLGPROC   lpDialogFunc,
            LPARAM    dwInitParam
        );
        '''
        hInstance, lpTemplateName, hWndParent, lpDialogFunc, dwInitParam = argv
        rv = self.get_handle()
        cw = self.get_char_width(ctx)
        if lpTemplateName:
            tname = self.read_mem_string(lpTemplateName, cw)
            argv[1] = tname

        return rv

    @apihook('CreateDialogIndirectParam', argc=5)
    def CreateDialogIndirectParam(self, emu, argv, ctx={}):
        '''
        HWND CreateDialogIndirectParam(
        HINSTANCE       hInstance,
        LPCDLGTEMPLATEA lpTemplate,
        HWND            hWndParent,
        DLGPROC         lpDialogFunc,
        LPARAM          dwInitParam
        );
        '''

        hnd, template, hnd_parent, func, param, = argv

        cb_args = (hnd_parent, windefs.WM_INITDIALOG, param, 0)
        self.setup_callback(func, cb_args, caller_argv=argv)
        return self.get_handle()

    @apihook('GetMenuInfo', argc=2)
    def GetMenuInfo(self, emu, argv, ctx={}):
        '''
        BOOL GetMenuInfo(
            HMENU,
            LPMENUINFO
        );
        '''
        return 1

    @apihook('GetProcessWindowStation', argc=0)
    def GetProcessWindowStation(self, emu, argv, ctx={}):
        '''
        HWINSTA GetProcessWindowStation();
        '''
        sta = self.sessman.get_current_station()
        return sta.get_handle()

    @apihook('LoadAccelerators', argc=2)
    def LoadAccelerators(self, emu, argv, ctx={}):
        '''
        HACCEL LoadAccelerators(
        HINSTANCE hInstance,
        LPCSTR    lpTableName
        );
        '''
        return self.get_handle()

    @apihook('IsWindowVisible', argc=1)
    def IsWindowVisible(self, emu, argv, ctx={}):
        '''
        BOOL IsWindowVisible(
        HWND hWnd
        );
        '''
        return True

    @apihook('BeginPaint', argc=2)
    def BeginPaint(self, emu, argv, ctx={}):
        '''
        HDC BeginPaint(
        HWND          hWnd,
        LPPAINTSTRUCT lpPaint
        );
        '''
        return self.get_handle()

    @apihook('LookupIconIdFromDirectory', argc=2)
    def LookupIconIdFromDirectory(self, emu, argv, ctx={}):
        '''
        int LookupIconIdFromDirectory(
        PBYTE presbits,
        BOOL  fIcon
        );
        '''
        return 1

    @apihook('GetActiveWindow', argc=0)
    def GetActiveWindow(self, emu, argv, ctx={}):
        '''
        HWND GetActiveWindow();
        '''
        return self.get_handle()

    @apihook('GetLastActivePopup', argc=1)
    def GetLastActivePopup(self, emu, argv, ctx={}):
        '''
        HWND GetLastActivePopup(
        HWND hWnd
        );
        '''
        hWnd, = argv
        return self.get_handle()

    @apihook('GetUserObjectInformation', argc=5)
    def GetUserObjectInformation(self, emu, argv, ctx={}):
        '''
        BOOL GetUserObjectInformation(
            HANDLE  hObj,
            int     nIndex,
            PVOID   pvInfo,
            DWORD   nLength,
            LPDWORD lpnLengthNeeded
        );
        '''
        obj, index, info, length, needed = argv

        if index == UOI_FLAGS:
            uoi = windefs.USEROBJECTFLAGS(emu.get_ptr_size())
            uoi.fInherit = 1
            uoi.dwFlags = 1

            if info:
                self.mem_write(info, uoi.get_bytes())

        return True

    @apihook('LoadIcon', argc=2)
    def LoadIcon(self, emu, argv, ctx={}):
        '''
        HICON LoadIcon(
            HINSTANCE hInstance,
            LPCSTR    lpIconName
        );
        '''
        inst, name, = argv

        if name not in (IDI_APPLICATION, IDI_ASTERISK, IDI_ERROR, IDI_EXCLAMATION, IDI_HAND,
                        IDI_INFORMATION, IDI_QUESTION, IDI_SHIELD, IDI_WARNING, IDI_WINLOGO):
            return 0
        return 1

    @apihook('GetRawInputDeviceList', argc=3)
    def GetRawInputDeviceList(self, emu, argv, ctx={}):
        """
        UINT GetRawInputDeviceList(
          PRAWINPUTDEVICELIST pRawInputDeviceList,
          PUINT               puiNumDevices,
          UINT                cbSize
        );
        """
        pRawInputDeviceList, puiNumDevices, cbSize = argv
        num_devices = 4
        self.mem_write(puiNumDevices, num_devices.to_bytes(4, 'little'))
        return num_devices

    @apihook('GetNextDlgTabItem', argc=3)
    def GetNextDlgTabItem(self, emu, argv, ctx={}):
        """
        HWND GetNextDlgTabItem(
          HWND hDlg,
          HWND hCtl,
          BOOL bPrevious
        );
        """
        return 0

    @apihook('GetCaretPos', argc=1)
    def GetCaretPos(self, emu, argv, ctx={}):
        """
        BOOL GetCaretPos(
          LPPOINT lpPoint
        );
        """
        lpPoint = argv[0]
        point = windef.POINT(emu.get_ptr_size())
        point.x = 0
        point.y = 0
        self.mem_write(lpPoint, self.get_bytes(point))
        return 1

    @apihook('GetMonitorInfo', argc=2)
    def GetMonitorInfo(self, emu, argv, ctx={}):
        """
        BOOL GetMonitorInfo(
          HMONITOR      hMonitor,
          LPMONITORINFO lpmi
        );
        """
        hMonitor, lpmi = argv
        mi = windef.MONITORINFO(emu.get_ptr_size())
        mi = self.mem_cast(mi, lpmi)
        # just a stub for now
        self.mem_write(lpmi, self.get_bytes(mi))
        return 1

    @apihook('EndPaint', argc=2)
    def EndPaint(self, emu, argv, ctx={}):
        """
        BOOL EndPaint(
          HWND              hWnd,
          const PAINTSTRUCT *lpPaint
        );
        """
        return 1

    @apihook('GetDlgCtrlID', argc=1)
    def GetDlgCtrlID(self, emu, argv, ctx={}):
        """
        int GetDlgCtrlID(
          HWND hWnd
        );
        """
        return 1

    @apihook('GetUpdateRect', argc=3)
    def GetUpdateRect(self, emu, argv, ctx={}):
        """
        BOOL GetUpdateRect(
          HWND   hWnd,
          LPRECT lpRect,
          BOOL   bErase
        );
        """
        return 0

    @apihook('GetAltTabInfo', argc=5)
    def GetAltTabInfo(self, emu, argv, ctx={}):
        """
        BOOL GetAltTabInfoA(
          HWND        hwnd,
          int         iItem,
          PALTTABINFO pati,
          LPSTR       pszItemText,
          UINT        cchItemText
        );
        """
        return 0

    @apihook('GetUpdateRgn', argc=3)
    def GetUpdateRgn(self, emu, argv, ctx={}):
        """
        int GetUpdateRgn(
          HWND hWnd,
          HRGN hRgn,
          BOOL bErase
        );
        """
        return 0

    @apihook('FlashWindow', argc=2)
    def FlashWindow(self, emu, argv, ctx={}):
        """
        BOOL FlashWindow(
          HWND hWnd,
          BOOL bInvert
        );
        """
        return 1

    @apihook('IsClipboardFormatAvailable', argc=1)
    def IsClipboardFormatAvailable(self, emu, argv, ctx={}):
        """
        BOOL IsClipboardFormatAvailable(
          UINT format
        );
        """
        return 0

    @apihook('IsWindow', argc=1)
    def IsWindow(self, emu, argv, ctx={}):
        """
        BOOL IsWindow(
            HWND hWnd
        );
        """
        hnd, = argv

        return True

    @apihook('EnableWindow', argc=2)
    def EnableWindow(self, emu, argv, ctx={}):
        """
        BOOL EnableWindow(
        HWND hWnd,
        BOOL bEnable
        );
        """
        hnd, bEnable = argv

        return False

    @apihook('CharLowerBuff', argc=2)
    def CharLowerBuff(self, emu, argv, ctx={}):
        """
        DWORD CharLowerBuffA(
            LPSTR lpsz,
            DWORD cchLength
        );
        """
        _str, cchLength = argv
        cw = self.get_char_width(ctx)
        val = self.read_mem_string(_str, cw, max_chars=cchLength)
        argv[0] = val
        argv[1] = cchLength
        self.write_mem_string(val.lower(), _str, cw)
        return cchLength

    @apihook('CharUpperBuff', argc=2)
    def CharUpperBuff(self, emu, argv, ctx={}):
        """
        DWORD CharUpperBuffA(
            LPSTR lpsz,
            DWORD cchLength
        );
        """
        _str, cchLength = argv
        cw = self.get_char_width(ctx)
        val = self.read_mem_string(_str, cw, max_chars=cchLength)
        argv[0] = val
        argv[1] = cchLength
        self.write_mem_string(val.upper(), _str, cw)
        return cchLength

    @apihook('CharLower', argc=1)
    def CharLower(self, emu, argv, ctx={}):
        """
        LPSTR CharLowerA(
            LPSTR lpsz
        );
        """
        _str, = argv
        cw = self.get_char_width(ctx)
        bits = _str.bit_length()
        if bits <= 16:
            if cw == 1:
                val = chr(_str).lower().encode('ascii')
            else:
                val = chr(_str).lower().encode('utf-16le')
            return int.from_bytes(val, byteorder='little')
        else:
            val = self.read_mem_string(_str, cw)
            self.write_mem_string(val.lower(), _str, cw)
            return _str

    @apihook('CharUpper', argc=1)
    def CharUpper(self, emu, argv, ctx={}):
        """
        LPSTR CharUpperA(
            LPSTR lpsz
        );
        """
        _str, = argv
        cw = self.get_char_width(ctx)
        bits = _str.bit_length()
        if bits <= 16:
            if cw == 1:
                val = chr(_str).upper().encode('ascii')
            else:
                val = chr(_str).upper().encode('utf-16le')
            return int.from_bytes(val, byteorder='little')
        else:
            val = self.read_mem_string(_str, cw)
            self.write_mem_string(val.upper(), _str, cw)
            return _str

    @apihook('SetTimer', argc=4)
    def SetTimer(self, emu, argv, ctx={}):
        """
        UINT_PTR SetTimer(
          HWND      hWnd,
          UINT_PTR  nIDEvent,
          UINT      uElapse,
          TIMERPROC lpTimerFunc
        );
        """
        self.timer_count += 1

        return self.get_handle()

    @apihook('KillTimer', argc=2)
    def KillTimer(self, emu, argv, ctx={}):
        """
        BOOL KillTimer(
          HWND     hWnd,
          UINT_PTR uIDEvent
        );
        """
        self.timer_count -= 1

        return True

    @apihook('OpenDesktop', argc=4)
    def OpenDesktop(self, emu, argv, ctx={}):
        """
        HDESK OpenDesktopA(
            LPCSTR      lpszDesktop,
            DWORD       dwFlags,
            BOOL        fInherit,
            ACCESS_MASK dwDesiredAccess
        );
        """
        lpszDesktop, dwFlags, fInherit, dwDesiredAccess = argv
        cw = self.get_char_width(ctx)
        desktop = self.read_mem_string(lpszDesktop, cw)
        argv[0] = desktop
        return self.get_handle()

    @apihook('SetThreadDesktop', argc=1)
    def SetThreadDesktop(self, emu, argv, ctx={}):
        """
        BOOL SetThreadDesktop(
            HDESK hDesktop
        );
        """
        return 0

    @apihook('GetKeyboardLayoutList', argc=2)
    def GetKeyboardLayoutList(self, emu, argv, ctx={}):
        """
        int GetKeyboardLayoutList(
          int nBuff,
          HKL *lpList
        );
        """
        nBuff, lpList = argv

        locale = 0x409      # English - United States
        self.mem_write(lpList, locale.to_bytes(2, 'little'))
        self.mem_write(lpList + 4, locale.to_bytes(2, 'little'))

        return 1



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import re
import json
import fnmatch
import traceback
from typing import List, Tuple, Dict

import binascii

import speakeasy.common as common
import speakeasy.winenv.arch as e_arch
from speakeasy.engines import unicorn_eng
from speakeasy.profiler import Profiler
import speakeasy.version as version
from speakeasy.memmgr import MemoryManager

from speakeasy.errors import EmuException

EMU_ENGINES = (
              ('unicorn', unicorn_eng.EmuEngine),
              )

WILDCARD_FLAG = bool
API_LEVEL = Tuple[Dict[str, List[common.ApiHook]], WILDCARD_FLAG]
MODULE_LEVEL = Tuple[Dict[str, API_LEVEL], WILDCARD_FLAG]


# Generic emulator class for binary code
class BinaryEmulator(MemoryManager):
    """
    Base class for emulating binaries
    """
    def __init__(self, config, logger=None):

        super(BinaryEmulator, self).__init__()

        self.stack_base = 0
        self.page_size = None
        self.inst_count = 0
        self.curr_instr_size = 0
        self.disasm_eng = None
        self.builtin_hooks_set = False
        self.emu_eng = None
        self.maps = []
        self.config = config
        self.hooks = {}

        self.profiler = Profiler()

        self.runtime = 0

        self.emu_version = self.get_emu_version()
        self.logger = logger

    def log_info(self, msg):
        if self.logger:
            self.logger.info(msg)

    def log_error(self, msg):
        if self.logger:
            self.logger.error(msg)

    def log_exception(self, msg):
        if self.logger:
            self.logger.exception(msg)

    def get_profiler(self):
        """
        Get the current event profiler object (if any)
        """
        return self.profiler

    def get_report(self):
        """
        Get the emulation report for all runs that were executed
        """
        if self.profiler:
            return self.profiler.get_report()

    def get_json_report(self):
        """
        Get the emulation report for all runs that were executed formatted as a JSON string
        """
        if self.profiler:
            return self.profiler.get_json_report()

    def _parse_config(self, config):
        """
        Parse the config to be used for emulation
        """
        if isinstance(config, str):
            config = json.loads(config)
        self.config = config

        _eng = config.get('emu_engine', '')
        for name, eng in EMU_ENGINES:
            if name.lower() == _eng.lower():
                self.emu_eng = eng()
        if not self.emu_eng:
            raise EmuException('Unsupported emulation engine: %s' % (_eng))

        self.osversion = config.get('os_ver', {})
        self.env = config.get('env', {})
        self.user_config = config.get('user', {})
        self.domain = config.get('domain')
        self.hostname = config.get('hostname')
        self.symlinks = config.get('symlinks', [])
        self.config_modules = config.get('modules', {})
        self.config_system_modules = self.config_modules.get('system_modules', [])
        self.config_processes = config.get('processes', [])
        self.config_user_modules = self.config_modules.get('user_modules', [])

        self.config_analysis = config.get('analysis', {})
        self.max_instructions = config.get('max_instructions', -1)
        self.timeout = config.get('timeout', 0)
        self.max_api_count = config.get('max_api_count', 5000)
        self.exceptions = config.get('exceptions', {})
        self.drive_config = config.get('drives', [])
        self.filesystem_config = config.get('filesystem', {})
        self.keep_memory_on_free = config.get('keep_memory_on_free', False)

        self.network_config = config.get('network', {})
        self.command_line = config.get('command_line', '')

    def get_emu_version(self):
        """
        Get the version of the emulator
        """
        return version.__version__

    def get_os_version(self):
        """
        Get version of the OS being emulated
        """
        return self.osversion

    def get_osver_string(self):
        """
        Get the human readable OS version string
        """
        osver = self.get_os_version()
        if osver:
            os_name = osver.get('name', '')
            major = osver.get('major')
            minor = osver.get('minor')
            if major is not None and minor is not None:
                verstr = '%s.%d_%d' % (os_name, major, minor)
                return verstr

    def get_domain(self):
        """
        Get domain of the machine being emulated
        """
        return self.domain

    def get_hostname(self):
        """
        Get hostname of the machine being emulated
        """
        return self.hostname

    def get_user(self):
        """
        Get the current emulated user's name
        """
        return self.user_config

    def sizeof(self, obj):
        """
        Get the size (in the emulation space) of the supplied object
        """
        return obj.sizeof()

    def get_bytes(self, obj):
        """
        Get the bytes represented in the emulation space of the supplied object
        """
        return obj.get_bytes()

    def stop(self):
        """
        Stop emulation completely
        """
        self.emu_eng.stop()
        if self.profiler:
            self.profiler.stop_run_clock()

    def start(self, addr, size):
        """
        Begin emulation
        """
        self.set_hooks()
        self._set_emu_hooks()
        if self.profiler:
            self.profiler.set_start_time()
        try:
            self.emu_eng.start(addr, timeout=self.timeout, count=self.max_instructions)
        except Exception:
            if self.profiler:
                self.profiler.log_error(traceback.format_exc())
            self.on_emu_complete()

    def get_network_config(self):
        """
        Get the network settings specified in the network section of the config file
        """
        return self.network_config

    def get_filesystem_config(self):
        """
        Get the filesystem settings specified in the filesystem section of the config file
        """
        return self.filesystem_config

    def get_drive_config(self):
        """
        Get the drive settings specified in the drives section of the config file
        """
        return self.drive_config

    def reg_write(self, reg, val):
        """
        Write a value to an emulated cpu register
        """
        if isinstance(reg, str):
            _reg = e_arch.REG_LOOKUP.get(reg.lower())
            if not _reg:
                raise EmuException('Invalid register access %s' % (reg))
            reg = _reg

        self.emu_eng.reg_write(reg, val)

    def reg_read(self, reg):
        """
        Read a value from an emulated cpu register
        """
        if isinstance(reg, str):
            _reg = e_arch.REG_LOOKUP.get(reg.lower())
            if not _reg:
                raise EmuException('Invalid register access %s' % (reg))
            reg = _reg

        return self.emu_eng.reg_read(reg)

    def set_hooks(self):
        """
        Set instruction level hooks
        """
        for ht in (common.HOOK_CODE, common.HOOK_MEM_READ, common.HOOK_MEM_WRITE,
                   common.HOOK_MEM_INVALID, common.HOOK_INTERRUPT):
            for hook in self.hooks.get(ht, []):
                if not hook.added:
                    hook.add()

    def _cs_disasm(self, mem, addr, fast=True):
        """
        Disassemble bytes using capstone
        """
        try:
            if fast:
                tu = [i for i in self.disasm_eng.disasm_lite(bytes(mem), addr)]
                address, size, mnem, oper = tu[0]
            else:
                return [i for i in self.disasm_eng.disasm(bytes(mem), addr)]
        except IndexError:
            raise EmuException("Failed to disasm at address: 0x%x" % (addr))

        op = '%s %s' % (mnem, oper)
        return ((mnem, oper, op))

    def disasm(self, mem, addr, fast=True):
        """
        Disassemble bytes at a specified address
        """
        return self._cs_disasm(mem, addr, fast=fast)

    def get_register_state(self):
        """
        Get the current state of registers from the emulator
        """
        regs = {}
        if e_arch.ARCH_X86 == self.get_arch():
            for name, reg in (('esp', e_arch.X86_REG_ESP),
                              ('ebp', e_arch.X86_REG_EBP),
                              ('eip', e_arch.X86_REG_EIP),
                              ('esi', e_arch.X86_REG_ESI),
                              ('edi', e_arch.X86_REG_EDI),
                              ('eax', e_arch.X86_REG_EAX),
                              ('ebx', e_arch.X86_REG_EBX),
                              ('ecx', e_arch.X86_REG_ECX),
                              ('edx', e_arch.X86_REG_EDX)):
                val = self.reg_read(reg)
                regs[name] = "{0:#0{1}x}".format(val, 2 + (self.get_ptr_size() * 2))
        elif e_arch.ARCH_AMD64 == self.get_arch():
            for name, reg in (('rsp', e_arch.AMD64_REG_RSP),
                              ('rbp', e_arch.AMD64_REG_RBP),
                              ('rip', e_arch.AMD64_REG_RIP),
                              ('rsi', e_arch.AMD64_REG_RSI),
                              ('rdi', e_arch.AMD64_REG_RDI),
                              ('rax', e_arch.AMD64_REG_RAX),
                              ('rbx', e_arch.AMD64_REG_RBX),
                              ('rcx', e_arch.AMD64_REG_RCX),
                              ('rdx', e_arch.AMD64_REG_RDX),
                              ('r8',  e_arch.AMD64_REG_R8),
                              ('r9',  e_arch.AMD64_REG_R9),
                              ('r10', e_arch.AMD64_REG_R10),
                              ('r11', e_arch.AMD64_REG_R11),
                              ('r12', e_arch.AMD64_REG_R12),
                              ('r13', e_arch.AMD64_REG_R13),
                              ('r14', e_arch.AMD64_REG_R14),
                              ('r15', e_arch.AMD64_REG_R15)):
                val = self.reg_read(reg)
                regs[name] = "{0:#0{1}x}".format(val, 2 + (self.get_ptr_size() * 2))
        return regs

    def get_disasm(self, addr, size, fast=True):
        """
        Get the disassembly from an address
        """
        return self.disasm(self.mem_read(addr, size), addr, fast)

    def set_func_args(self, stack_addr, ret_addr, *args, home_space=True):
        """
        Set the arguments before an emulated function call. This is how we pass
        arguments to a function when calling it through the emulator.
        """
        curr_sp = stack_addr - self.ptr_size
        nargs = len(args)

        if self.get_arch() == e_arch.ARCH_X86:
            sp = e_arch.X86_REG_ESP
        elif self.get_arch() == e_arch.ARCH_AMD64:
            sp = e_arch.AMD64_REG_RSP
            i = 0
            for i, r in enumerate((e_arch.AMD64_REG_RCX, e_arch.AMD64_REG_RDX,
                                   e_arch.AMD64_REG_R8, e_arch.AMD64_REG_R9)):
                if nargs == 0:
                    break
                self.reg_write(r, args[i])
                nargs -= 1
            # Set the stack home space
            if home_space:
                curr_sp -= 0x20
            self.reg_write(sp, curr_sp)
        else:
            raise EmuException('Unsupported architecture')

        if nargs > 0:
            for arg in args[-nargs:][::-1]:
                a = arg.to_bytes(self.ptr_size, byteorder='little')

                self.mem_write(curr_sp, a)
                self.reg_write(sp, curr_sp)
                curr_sp -= self.ptr_size

        # Set the return address
        r = ret_addr.to_bytes(self.ptr_size, byteorder='little')
        self.mem_write(curr_sp, r)
        self.reg_write(sp, curr_sp)

    def get_func_argv(self, callconv, argc):
        """
        Get the arguments for a function given the supplied calling convention
        """
        argv = []
        ptr_size = self.get_ptr_size()
        arch = self.get_arch()
        nargs = argc
        endian = 'little'

        # Handle calling conventions using floats
        if arch in (e_arch.ARCH_X86, e_arch.ARCH_AMD64):
            if callconv == e_arch.CALL_CONV_FLOAT:

                for i, r in enumerate((e_arch.X86_REG_XMM0,
                                       e_arch.X86_REG_XMM1,
                                       e_arch.X86_REG_XMM2,
                                       e_arch.X86_REG_XMM3)):
                    if nargs == 0:
                        break
                    val = self.reg_read(r)
                    argv.append(val)
                    nargs -= 1

        if arch == e_arch.ARCH_X86:
            sp = self.reg_read(e_arch.X86_REG_ESP)
            if callconv == e_arch.CALL_CONV_FASTCALL:
                if nargs >= 2:
                    argv.append(self.reg_read(e_arch.X86_REG_ECX))
                    argv.append(self.reg_read(e_arch.X86_REG_EDX))
                    nargs -= 2
                elif nargs == 1:
                    argv.append(self.reg_read(e_arch.X86_REG_ECX))
                    nargs -= 1
        elif arch == e_arch.ARCH_AMD64:
            sp = self.reg_read(e_arch.AMD64_REG_RSP)
            sp += 0x20

            for i, r in enumerate((e_arch.AMD64_REG_RCX, e_arch.AMD64_REG_RDX,
                                   e_arch.AMD64_REG_R8, e_arch.AMD64_REG_R9)):
                if nargs == 0:
                    break
                val = self.reg_read(r)
                argv.append(val)
                nargs -= 1
        else:
            raise EmuException('Unsupported architecture')

        # Skip past the saved ret addr
        sp += ptr_size
        for i in range(nargs):
            ptr = self.mem_read(sp, ptr_size)
            argv.append(int.from_bytes(ptr, endian))
            sp += ptr_size

        return argv

    def do_call_return(self, argc, ret_addr=None, ret_value=None, conv=None):
        """
        Set the emulation state after a call has completed
        """
        if self.get_arch() == e_arch.ARCH_X86:
            sp = e_arch.X86_REG_ESP
            pc = e_arch.X86_REG_EIP
            rr = e_arch.X86_REG_EAX
        elif self.get_arch() == e_arch.ARCH_AMD64:
            sp = e_arch.AMD64_REG_RSP
            pc = e_arch.AMD64_REG_RIP
            rr = e_arch.AMD64_REG_RAX
        else:
            raise EmuException('Unsupported architecture')

        if conv == e_arch.CALL_CONV_FLOAT:
            rr = e_arch.X86_REG_XMM0

        stk_ptr = self.reg_read(sp)

        if ret_addr:
            self.reg_write(sp, stk_ptr + self.ptr_size)
            self.reg_write(pc, ret_addr)
        if ret_value is not None:
            self.reg_write(rr, ret_value)

        # Cleanup the stack
        if conv == e_arch.CALL_CONV_CDECL:
            # If cdecl, the emu engine will clean the stack
            pass
        elif conv == e_arch.CALL_CONV_FASTCALL:
            if self.get_arch() == e_arch.ARCH_X86:
                if argc > 2:
                    self.clean_stack_args(argc - 2)
        else:
            self.clean_stack_args(argc)

    def get_ret_address(self):
        """
        Get the return address from the stack
        """

        endian = 'little'

        sp = self.get_stack_ptr()
        ret = self.mem_read(sp, self.ptr_size)
        ret = int.from_bytes(ret, endian)
        return ret

    def push_stack(self, val):
        """
        Put a value on the stack and adjust the stack pointer
        """
        endian = 'little'
        sp = self.get_stack_ptr()
        bval = val.to_bytes(self.ptr_size, endian)
        sp -= self.ptr_size
        self.mem_write(sp, bval)
        self.set_stack_ptr(sp)
        return val

    def pop_stack(self):
        """
        Get value from the stack and adjust the stack pointer
        """
        endian = 'little'
        sp = self.get_stack_ptr()
        val = self.mem_read(sp, self.ptr_size)
        val = int.from_bytes(val, endian)
        sp += self.ptr_size
        self.set_stack_ptr(sp)
        return val

    def get_stack_ptr(self):
        """
        Get the current address of the stack pointer
        """
        if self.get_arch() == e_arch.ARCH_X86:
            sp = self.reg_read(e_arch.X86_REG_ESP)
        elif self.get_arch() == e_arch.ARCH_AMD64:
            sp = self.reg_read(e_arch.AMD64_REG_RSP)
        return sp

    def set_stack_ptr(self, addr):
        """
        Set the current address of the stack pointer
        """
        if self.get_arch() == e_arch.ARCH_X86:
            self.reg_write(e_arch.X86_REG_ESP, addr)
        elif self.get_arch() == e_arch.ARCH_AMD64:
            self.reg_write(e_arch.AMD64_REG_RSP, addr)

    def format_stack(self, num_ptrs):
        """
        Get the stack and format it for display
        """
        out = []
        sp = self.get_stack_ptr()
        for i in range(num_ptrs):
            try:
                ptr = self.mem_read(sp, self.get_ptr_size())
            except Exception:
                return out
            ptr = int.from_bytes(ptr, 'little')
            tag = self.get_address_tag(ptr)
            out.append((sp, ptr, tag))
            sp += self.get_ptr_size()
        return out

    def print_stack(self, num_ptrs):
        """
        This a debug function used to print the current stack state
        """
        ptrs = self.format_stack(num_ptrs)
        print('Stack:')
        print('***********************')
        for p in ptrs:
            sp, ptr, tag = p
            if tag:
                fmt = 'sp=0x%x:\t0x%x\t->\t%s' % (sp, ptr, tag)
            else:
                fmt = 'sp=0x%x:\t0x%x\t' % (sp, ptr)

            print(fmt.expandtabs(5))
            sp += self.get_ptr_size()

    def get_stack_trace(self, num_ptrs=16):
        """
        Get the current stack state
        """
        trace = []
        sp = self.get_stack_ptr()
        try:
            for i in range(num_ptrs):
                ptr = self.mem_read(sp, self.get_ptr_size())
                ptr = int.from_bytes(ptr, 'little')
                tag = self.get_address_tag(ptr)
                fmt = "{0:#0{1}x}".format(ptr, 2 + (self.get_ptr_size() * 2))
                sp_off = "{0:#0{1}x}".format(i * self.get_ptr_size(), 2 * 2)
                if not tag:
                    entry = 'sp+%s: %s' % (sp_off, fmt)
                else:
                    entry = 'sp+%s: %s -> %s' % (sp_off, fmt, tag)
                trace.append(entry)
                sp += self.get_ptr_size()
        finally:
            return trace

    def get_pc(self):
        """
        Get the value of the current program counter
        """
        if self.get_arch() == e_arch.ARCH_X86:
            pc = self.reg_read(e_arch.X86_REG_EIP)
        elif self.get_arch() == e_arch.ARCH_AMD64:
            pc = self.reg_read(e_arch.AMD64_REG_RIP)
        else:
            raise EmuException('Unsupported architecture')
        return pc

    def set_pc(self, addr):
        """
        Set the value of the current program counter
        """
        if self.get_arch() == e_arch.ARCH_X86:
            self.reg_write(e_arch.X86_REG_EIP, addr)
        elif self.get_arch() == e_arch.ARCH_AMD64:
            self.reg_write(e_arch.AMD64_REG_RIP, addr)
        else:
            raise EmuException('Unsupported architecture')

    def get_return_val(self):
        """
        Get the current value in the return register
        """
        if self.get_arch() == e_arch.ARCH_X86:
            val = self.reg_read(e_arch.X86_REG_EAX)
        elif self.get_arch() == e_arch.ARCH_AMD64:
            val = self.reg_read(e_arch.AMD64_REG_RAX)
        else:
            raise EmuException('Unsupported architecture')
        return val

    def reset_stack(self, base):
        """
        Reset stack to the supplied base address
        """
        arch = self.get_arch()
        ptr = base

        if arch == e_arch.ARCH_X86:
            self.reg_write(e_arch.X86_REG_ESP, base)
            self.reg_write(e_arch.X86_REG_EBP, base)
        elif arch == e_arch.ARCH_AMD64:
            # Save room for the "home space"
            ptr -= self.ptr_size * 5
            self.reg_write(e_arch.AMD64_REG_RSP, ptr)
            self.reg_write(e_arch.AMD64_REG_RBP, ptr)

        return base, ptr

    def alloc_stack(self, size):
        """
        Allocate memory to use for the program stack
        """
        # Allocate memory for our stack
        # Stack grows down
        chunk = self.get_valid_ranges(size, addr=0x1200000)
        addr, block_size = chunk
        self.mem_map(block_size, base=addr, tag='emu.stack')

        base = addr + block_size
        self.mem_reserve(size, base=base)

        base, ptr = self.reset_stack(base)

        return base, ptr

    def clean_stack_args(self, argc):
        """
        Adjust the stack for arguments that were supplied
        """
        ptr_size = self.get_ptr_size()
        arch = self.get_arch()

        if argc == 0:
            return

        if arch == e_arch.ARCH_X86:
            sp = self.reg_read(e_arch.X86_REG_ESP)
            sp += (ptr_size * argc)
            self.reg_write(e_arch.X86_REG_ESP, sp)

        elif arch == e_arch.ARCH_AMD64:
            return
        else:
            raise EmuException('Unsupported architecture')

    def get_arch(self):
        """
        Get the current emulated architecture
        """
        return self.arch

    def get_arch_name(self):
        """
        Get the name of current emulated architecture
        """
        if self.arch == e_arch.ARCH_AMD64:
            return 'amd64'
        elif self.arch == e_arch.ARCH_X86:
            return 'x86'
        return ''

    def eval_emu_var(self):
        """
        Used to expand variables supplied in the emulator config file. This
        might be useful for accessing files that are a relative path of the
        speakeasy package.
        For example:
            $ROOT$: This variable corresponds to the package root for speakeasy
        """

    def read_mem_string(self, address, width=1, max_chars=0):
        """
        Read a string from emulated memory
        """
        char = b'\xFF'
        string = b''
        i = 0

        if width == 1:
            decode = 'latin1'
        elif width == 2:
            decode = 'utf-16le'
        else:
            raise ValueError('Invalid string encoding')

        while int.from_bytes(char, 'little') != 0:
            if max_chars and i >= max_chars:
                break
            char = self.mem_read(address, width)

            string += char
            address += width
            i += 1

        try:
            dec = string.decode(decode, 'ignore').replace('\x00', '')
        except Exception:
            dec = string.replace(b'\x00', b'')
        return dec

    def mem_string_len(self, address, width=1):
        """
        Get the length of a string from emulated memory
        """
        slen = -1
        char = b'\xFF'

        while int.from_bytes(char, 'little') != 0:
            char = self.mem_read(address, width)
            address += width
            slen += 1
        return slen

    def get_ansi_strings(self, data, min_len=4):
        """
        Get all ansi strings from a supplied memory blob
        """
        astrs = []
        pat = b'[\x20-\x7f]{%d,}' % (min_len)
        res = re.compile(pat)
        hits = res.findall(data)
        offset = 0
        for s in hits:
            try:
                offset = data.find(s, offset)
                s = s.decode('utf-8')
                astrs.append((offset, s))
                offset += 1
            except UnicodeDecodeError:
                continue
        return astrs

    def get_unicode_strings(self, data, min_len=4):
        """
        Get all unicode strings from a supplied memory blob
        """
        wstrs = []
        pat = b'(?:[\x20-\x7f]\x00){%d,}' % (min_len)
        res = re.compile(pat)
        hits = res.findall(data)
        offset = 0
        for ws in hits:
            try:
                offset = data.find(ws, offset)
                ws = ws.decode('utf-16le')
                wstrs.append((offset, ws))
                offset += 1
            except UnicodeDecodeError:
                continue
        return wstrs

    def mem_copy(self, dst, src, n):
        """
        Copy bytes from one emulated address to another
        """
        sbytes = self.mem_read(src, n)
        self.mem_write(dst, sbytes)
        return n

    def write_mem_string(self, string, address, width=1):
        """
        Write string data to an emulated memory address
        """

        if width == 1:
            encode = 'utf-8'
        elif width == 2:
            encode = 'utf-16le'
        else:
            raise ValueError('Invalid string encoding')

        enc_str = string.encode(encode)
        self.mem_write(address, enc_str)

    def read_ptr(self, address):
        val = self.mem_read(address, self.ptr_size)
        return int.from_bytes(val, 'little')

    def write_ptr(self, address, val):
        self.mem_write(address, val.to_bytes(self.ptr_size, 'little'))

    def get_ptr_size(self):
        """
        Get the pointer size of the current emulation state
        """
        return self.ptr_size

    def get_mem_strings(self):
        """
        Get ansi and unicode strings from emulated memory
        """
        tgt_tag_prefixes = ('emu.stack', 'api')
        ansi_strings = []
        unicode_strings = []
        ret_ansi = []
        ret_unicode = []

        for mmap in self.get_mem_maps():
            tag = mmap.get_tag()
            if tag and tag.startswith(tgt_tag_prefixes) and tag != self.input.get('mem_tag'):
                data = self.mem_read(mmap.get_base(), mmap.get_size()-1)
                ansi_strings += self.get_ansi_strings(data)
                unicode_strings += self.get_unicode_strings(data)

        [ret_ansi.append(a) for a in ansi_strings if a not in ret_ansi]
        [ret_unicode.append(a) for a in unicode_strings if a not in ret_unicode]

        return (ret_ansi, ret_unicode)

    def set_ptr_size(self, arch):
        """
        Set the current pointer size used in the emulator
        """
        if arch == e_arch.ARCH_AMD64:
            self.ptr_size = 8
        elif arch == e_arch.ARCH_X86:
            self.ptr_size = 4
        else:
            raise EmuException('Unsupported architecture')

    def get_module_from_addr(self, addr):
        """
        If the supplied address belongs to a module, return it
        """
        for mod in self.modules:
            base, size = mod[1]
            if addr >= base and addr <= base + size:
                return mod[0]
        return None

    def get_api_hooks(self, mod_name, func_name) -> List[common.ApiHook]:
        """
        If an API hook has been set, return it here
        """

        mod_name = mod_name.lower()
        func_name = func_name.lower()
        try:
            hook_struct, wildcard_module = self.hooks[common.HOOK_API]
        except KeyError:
            return []
        try:
            modules = [hook_struct[mod_name]]
        except KeyError:
            modules = []
        if wildcard_module:
            for module_name_saved, value in hook_struct.items():
                if fnmatch.fnmatch(mod_name, module_name_saved) and mod_name != module_name_saved:
                    modules.append(value)
        user_hooks = []
        for module in modules:
            hooks, wildcard_api = module
            try:
                user_hooks.extend(hooks[func_name])
            except KeyError:
                pass
            if wildcard_api:
                for func_name_saved, list_of_hooks in hooks.items():
                    if (fnmatch.fnmatch(func_name, func_name_saved) and
                       func_name != func_name_saved):
                        user_hooks.extend(list_of_hooks)
        return user_hooks

    def add_api_hook(self, cb, module='', api_name='', argc=0, call_conv=None,
                     emu=None) -> common.ApiHook:
        """
        Add an API level hook (e.g. kernel32.CreateFile) here
        """
        module = module.lower()
        api_name = api_name.lower()

        wildcard_module, wildcard_api = False, False
        for wc in ['?', '*', '[', ']']:
            if wc in module:
                wildcard_module = True
            if wc in api_name:
                wildcard_api = True

        if not emu:
            emu = self
        hook = common.ApiHook(emu, self.emu_eng, cb, module, api_name, argc, call_conv)
        _hooks: MODULE_LEVEL = self.hooks.get(common.HOOK_API)

        api_dictionary = (
            {
              api_name: [hook]
            },
            wildcard_api)
        if not _hooks:
            # First addition
            obj = ({module: api_dictionary}, wildcard_module)
        else:
            module_dict, previous_wildcard_module = _hooks
            try:
                api_dict, previous_wildcard_api = module_dict[module]
            except KeyError:
                # The module asked is not present, so we just add the api dictionary
                module_dict[module] = api_dictionary
            else:
                # The module asked is present, so we can just add the hook
                api_dict.setdefault(api_name, []).append(hook)
                module_dict[module] = (api_dict, previous_wildcard_api | wildcard_api)
            obj = (module_dict, previous_wildcard_module | wildcard_module)
        self.hooks.update({common.HOOK_API: obj})
        return hook

    def add_code_hook(self, cb, begin=1, end=0, ctx={}, emu=None):
        """
        Add a hook that will fire for every CPU instruction
        """
        hl = self.hooks.get(common.HOOK_CODE, [])
        if not emu:
            emu = self
        hook = common.CodeHook(self, self.emu_eng, cb, begin, end, ctx)
        if not hl:
            self.hooks.update({common.HOOK_CODE: [hook, ]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook

    def _dynamic_code_cb(self, emu, addr, size, ctx={}):
        """
        Call all subscribers that want callbacks dynamic code callbacks
        """

        profiler = self.get_profiler()
        mm = self.get_address_map(addr)
        if profiler:
            run = self.get_current_run()
            profiler.log_dyn_code(run, mm.get_tag(), mm.get_base(), mm.get_size())

        for h in self.hooks.get(common.HOOK_DYN_CODE, []):
            h.cb(mm)

        # Delete the code hook that got us here
        if ctx and isinstance(ctx, dict):
            h = ctx.get('_delete_hook')
            if h:
                h.disable()

    def _set_dyn_code_hook(self, addr, size, ctx={}):
        """
        Set the top level dispatch hook for dynamic code execution
        """
        max_hook_size = 0x10
        if size > max_hook_size:
            size = max_hook_size

        ch = self.add_code_hook(cb=self._dynamic_code_cb, begin=addr, end=addr + size, ctx=ctx)
        ctx.update({'_delete_hook': ch})

    def add_dyn_code_hook(self, cb, ctx=[], emu=None):
        """
        Add a hook that will fire when dynamically generated/copied code is executed
        """
        if not emu:
            emu = self
        hl = self.hooks.get(common.HOOK_DYN_CODE, [])

        hook = common.DynCodeHook(emu, self.emu_eng, cb, ctx)
        if not hl:
            self.hooks.update({common.HOOK_DYN_CODE: [hook, ]})
        else:
            hl.insert(0, hook)

        return hook

    def add_mem_read_hook(self, cb, begin=1, end=0, emu=None):
        """
        Add a hook that will fire for memory reads
        """
        if not emu:
            emu = self
        hook = common.ReadMemHook(emu, self.emu_eng, cb, begin, end)
        hl = self.hooks.get(common.HOOK_MEM_READ)
        if not hl:
            self.hooks.update({common.HOOK_MEM_READ: [hook, ]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook

    def add_mem_write_hook(self, cb, begin=1, end=0, emu=None):
        """
        Add a hook that will fire for memory writes
        """
        if not emu:
            emu = self
        hook = common.WriteMemHook(emu, self.emu_eng, cb, begin, end)
        hl = self.hooks.get(common.HOOK_MEM_WRITE)
        if not hl:
            self.hooks.update({common.HOOK_MEM_WRITE: [hook, ]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook

    def add_mem_map_hook(self, cb, begin=1, end=0, emu=None):
        """
        Add a hook that will fire for memory maps
        """
        if not emu:
            emu = self
        hook = common.MapMemHook(emu, self.emu_eng, cb, begin, end)
        hl = self.hooks.get(common.HOOK_MEM_MAP)
        if not hl:
            self.hooks.update({common.HOOK_MEM_MAP: [hook, ]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook

    def _hook_mem_invalid_dispatch(self, emu, access, address, size, value, ctx):
        """
        This handler will dispatch other invalid memory hooks
        """
        hl = self.hooks.get(common.HOOK_MEM_INVALID, [])

        rv = True
        for mem_access_hook in hl[:-1]:
            if mem_access_hook.enabled:
                rv = mem_access_hook.cb(emu, access, address, size, value, ctx)
                if rv is False:
                    break
        return rv

    def add_mem_invalid_hook(self, cb, emu=None):
        """
        Add a hook that will fire for invalid memory access
        """
        hook = common.InvalidMemHook(self, self.emu_eng, cb, native_hook=False)
        hl = self.hooks.get(common.HOOK_MEM_INVALID)
        if not emu:
            emu = self
        if not hl:
            dispatch_hook = common.InvalidMemHook(emu, self.emu_eng,
                                                  self._hook_mem_invalid_dispatch,
                                                  native_hook=True)
            if self.emu_eng:
                dispatch_hook.add()

            self.hooks.update({common.HOOK_MEM_INVALID: [hook, dispatch_hook]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook

    def add_interrupt_hook(self, cb, ctx=[], emu=None):
        """
        Add a hook that will fire for software interrupts
        """
        if not emu:
            emu = self
        hook = common.InterruptHook(emu, self.emu_eng, cb, ctx=[])
        hl = self.hooks.get(common.HOOK_INTERRUPT)
        if not hl:
            self.hooks.update({common.HOOK_INTERRUPT: [hook, ]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook

    def add_instruction_hook(self, cb, begin=1, end=0, ctx=[], emu=None, insn=None):
        """
        Add a hook that will fire for IN, SYSCALL, or SYSENTER instructions
        """
        if not emu:
            emu = self
        hook = common.InstructionHook(emu, self.emu_eng, cb, ctx=[], insn=insn)
        hl = self.hooks.get(common.HOOK_INSN)
        if not hl:
            self.hooks.update({common.HOOK_INSN: [hook, ]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook

    def add_invalid_instruction_hook(self, cb, ctx=[], emu=None):
        if not emu:
            emu = self

        hook = common.InvalidInstructionHook(emu, self.emu_eng, cb, ctx=[])
        hl = self.hooks.get(common.HOOK_INSN_INVALID)

        if not hl:
            self.hooks.update({common.HOOK_INSN_INVALID: [hook, ]})
        else:
            hl.insert(0, hook)

        if self.emu_eng:
            hook.add()

        return hook



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.nt.ddk as ddk
from speakeasy.winenv.api import api


class Usbd(api.ApiHandler):
    """
    Implements the USB stack driver (USBD.sys)
    """

    name = 'usbd'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    class ApiEmuError(Exception):
        pass

    def __init__(self, emu):

        super(Usbd, self).__init__(emu)

        self.funcs = {}
        self.data = {}

        super(Usbd, self).__get_hook_attrs__(self)

    @apihook('USBD_ValidateConfigurationDescriptor', argc=5)
    def USBD_ValidateConfigurationDescriptor(self, emu, argv, ctx={}):
        """
        USBD_STATUS USBD_ValidateConfigurationDescriptor(
          PUSB_CONFIGURATION_DESCRIPTOR ConfigDesc,
          ULONG                         BufferLength,
          USHORT                        Level,
          PUCHAR                        *Offset,
          ULONG                         Tag
        );
        """
        rv = ddk.STATUS_SUCCESS
        ConfigDesc, BufferLength, Level, Offset, Tag = argv

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch
import speakeasy.winenv.defs.windows.mpr as mpr

from .. import api


class Mpr(api.ApiHandler):

    name = 'mpr'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Mpr, self).__init__(emu)
        super(Mpr, self).__get_hook_attrs__(self)

    @apihook('WNetOpenEnum', argc=5, conv=_arch.CALL_CONV_STDCALL)
    def WNetOpenEnum(self, emu, argv, ctx={}):
        """
        DWORD WNetOpenEnum(
          DWORD          dwScope,
          DWORD          dwType,
          DWORD          dwUsage,
          LPNETRESOURCEW lpNetResource,
          LPHANDLE       lphEnum
        );
        """
        dwScope, dwType, dwUsage, lpNetResource, lphEnum = argv

        scope = mpr.get_define_int(dwScope, 'RESOURCE_')
        if scope:
            argv[0] = scope

        type = mpr.get_define_int(dwType, 'RESOURCETYPE_')
        if type:
            argv[1] = type

        usage = mpr.get_define_int(dwUsage, 'RESOURCEUSAGE_')
        if usage:
            argv[2] = usage

        return mpr.ERROR_NO_NETWORK

    @apihook('WNetEnumResource', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WNetEnumResource(self, emu, argv, ctx={}):
        """
        DWORD WNetEnumResourceA(
          HANDLE  hEnum,
          LPDWORD lpcCount,
          LPVOID  lpBuffer,
          LPDWORD lpBufferSize
        );
        """
        return mpr.ERROR_NO_NETWORK

    @apihook('WNetAddConnection2', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def WNetAddConnection2(self, emu, argv, ctx={}):
        """
        DWORD WNetAddConnection2W(
          LPNETRESOURCEW lpNetResource,
          LPCWSTR        lpPassword,
          LPCWSTR        lpUserName,
          DWORD          dwFlags
        );
        """
        return mpr.ERROR_NO_NETWORK

    @apihook('WNetGetConnection', argc=3, conv=_arch.CALL_CONV_STDCALL)
    def WNetGetConnection(self, emu, argv, ctx={}):
        """
        DWORD WNetGetConnectionA(
          LPCSTR  lpLocalName,
          LPSTR   lpRemoteName,
          LPDWORD lpnLength
        );
        """
        lpLocalName, lpRemoteName, lpnLength = argv

        cw = self.get_char_width(ctx)

        local_name = self.read_mem_string(lpLocalName, cw)
        if local_name:
            argv[0] = local_name

        return mpr.ERROR_NO_NETWORK



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import ctypes as ct

from speakeasy.struct import EmuStruct, Ptr
import speakeasy.winenv.defs.windows.windows as windefs

from .. import api

DNS_TYPE_TEXT = 0x0010


class _DnsRecord(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.pNext = Ptr
        self.pName = Ptr
        self.wType = ct.c_uint16
        self.wDataLength = ct.c_uint16
        self.Flags = ct.c_uint32
        self.dwTtl = ct.c_uint32
        self.dwReserved = ct.c_uint32


class DNS_TXT_DATA(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.dwStringCount = Ptr
        self.pStringArray = Ptr


class DnsApi(api.ApiHandler):

    name = 'dnsapi'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(DnsApi, self).__init__(emu)
        self.funcs = {}
        self.data = {}
        super(DnsApi, self).__get_hook_attrs__(self)
        self.netman = emu.get_network_manager()
        self.names = {}

    @apihook('DnsQuery_', argc=6)
    def DnsQuery_(self, emu, argv, ctx={}):
        """
        DNS_STATUS DnsQuery_A(
            PCSTR       pszName,
            WORD        wType,
            DWORD       Options,
            PVOID       pExtra,
            PDNS_RECORD *ppQueryResults,
            PVOID       *pReserved
        );
        """

        pszName, wType, Options, pExtra, ppQueryResults, pReserved = argv
        rv = windefs.ERROR_INVALID_PARAMETER
        rr = None

        cw = self.get_char_width(ctx)
        if pszName:
            name = self.read_mem_string(pszName, cw)
            ip = self.netman.name_lookup(name)
            self.log_dns(name, ip)

            rec = _DnsRecord(emu.get_ptr_size())
            rec.pName = pszName
            rec.wType = wType
            if wType == DNS_TYPE_TEXT:
                argv[1] = 'DNS_TYPE_TEXT'

                text = self.netman.get_dns_txt(name)
                if not text:
                    text = b'\x00'*12

                ts = DNS_TXT_DATA(emu.get_ptr_size())
                size = len(text) + ts.sizeof()

                rr = self.mem_alloc(rec.sizeof() + size, tag='api.DnsQuery._DnsRecord')
                ts.dwStringCount = 1
                ts.pStringArray = rr + rec.sizeof() + ts.sizeof()
                rec.wDataLength = size

                self.mem_write(rr, rec.get_bytes() + ts.get_bytes())
                self.mem_write(ts.pStringArray, text)

            if ppQueryResults and rr:
                out = rr.to_bytes(self.get_ptr_size(), 'little')
                self.mem_write(ppQueryResults, out)
                rv = windefs.ERROR_SUCCESS

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import ctypes as ct

import speakeasy.winenv.defs.usb as usbdefs
from speakeasy.struct import EmuStruct, Ptr, EmuUnion, Enum


WdfUsbTargetDeviceSelectConfigType = Enum()
WdfUsbTargetDeviceSelectConfigType.WdfUsbTargetDeviceSelectConfigTypeInvalid = 0
WdfUsbTargetDeviceSelectConfigType.WdfUsbTargetDeviceSelectConfigTypeDeconfig = 1
WdfUsbTargetDeviceSelectConfigType.WdfUsbTargetDeviceSelectConfigTypeSingleInterface = 2
WdfUsbTargetDeviceSelectConfigType.WdfUsbTargetDeviceSelectConfigTypeMultiInterface = 3
WdfUsbTargetDeviceSelectConfigType.WdfUsbTargetDeviceSelectConfigTypeInterfacesPairs = 4
WdfUsbTargetDeviceSelectConfigType.WdfUsbTargetDeviceSelectConfigTypeInterfacesDescriptor = 5
WdfUsbTargetDeviceSelectConfigType.WdfUsbTargetDeviceSelectConfigTypeUrb = 6

WdfUsbTargetDeviceSelectSettingType = Enum()
WdfUsbTargetDeviceSelectSettingType.WdfUsbInterfaceSelectSettingTypeDescriptor = 0x10
WdfUsbTargetDeviceSelectSettingType.WdfUsbInterfaceSelectSettingTypeSetting = 0x11
WdfUsbTargetDeviceSelectSettingType.WdfUsbInterfaceSelectSettingTypeUrb = 0x12

WDF_USB_PIPE_TYPE = Enum()
WDF_USB_PIPE_TYPE.WdfUsbPipeTypeInvalid = 0
WDF_USB_PIPE_TYPE.WdfUsbPipeTypeControl = 1
WDF_USB_PIPE_TYPE.WdfUsbPipeTypeIsochronous = 2
WDF_USB_PIPE_TYPE.WdfUsbPipeTypeBulk = 3
WDF_USB_PIPE_TYPE.WdfUsbPipeTypeInterrupt = 4

WDF_USB_DEVICE_TRAIT_SELF_POWERED = 1
WDF_USB_DEVICE_TRAIT_REMOTE_WAKE_CAPABLE = 2
WDF_USB_DEVICE_TRAIT_AT_HIGH_SPEED = 4


class WDF_VERSION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Major = ct.c_uint32
        self.Minor = ct.c_uint32
        self.Build = ct.c_uint32


class WDF_BIND_INFO(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.Component = Ptr
        self.Version = WDF_VERSION
        self.FuncCount = ct.c_uint32
        self.FuncTable = Ptr
        self.Module = Ptr


class WDF_USB_DEVICE_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.UsbdVersionInformation = usbdefs.USBD_VERSION_INFORMATION
        self.HcdPortCapabilities = ct.c_uint32
        self.Traits = ct.c_uint32


class WDF_DRIVER_CONFIG(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.EvtDriverDeviceAdd = Ptr
        self.EvtDriverUnload = Ptr
        self.DriverInitFlags = ct.c_uint32
        self.DriverPoolTag = ct.c_uint32


class WDF_COMPONENT_GLOBALS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Data = ct.c_uint8 * 0x100


class WDF_TYPED_CONTEXT_WORKER(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Data = ct.c_uint8 * 0x100


class WDF_USB_PIPE_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.MaximumPacketSize = ct.c_uint32
        self.EndpointAddress = ct.c_uint8
        self.Interval = ct.c_uint8
        self.SettingIndex = ct.c_uint8
        self.PipeType = ct.c_uint32
        self.MaximumTransferSize = ct.c_uint32


class WDF_PNPPOWER_EVENT_CALLBACKS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.EvtDeviceD0Entry = Ptr
        self.EvtDeviceD0EntryPostInterruptsEnabled = Ptr
        self.EvtDeviceD0Exit = Ptr
        self.EvtDeviceD0ExitPreInterruptsDisabled = Ptr
        self.EvtDevicePrepareHardware = Ptr
        self.EvtDeviceReleaseHardware = Ptr
        self.EvtDeviceSelfManagedIoCleanup = Ptr
        self.EvtDeviceSelfManagedIoFlush = Ptr
        self.EvtDeviceSelfManagedIoInit = Ptr
        self.EvtDeviceSelfManagedIoSuspend = Ptr
        self.EvtDeviceSelfManagedIoRestart = Ptr
        self.EvtDeviceSurpriseRemoval = Ptr
        self.EvtDeviceQueryRemove = Ptr
        self.EvtDeviceQueryStop = Ptr
        self.EvtDeviceUsageNotification = Ptr
        self.EvtDeviceRelationsQuery = Ptr
        self.EvtDeviceUsageNotificationEx = Ptr


class _InterfaceDescriptor(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.InterfaceDescriptor = Ptr


class Interface(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.SettingIndex = ct.c_uint8


class InterfaceUrb(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Urb = Ptr


class InterfaceTypes(EmuUnion):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Descriptor = _InterfaceDescriptor(ptr_size).get_cstruct()
        self.Interface = Interface(ptr_size).get_cstruct()
        self.Urb = InterfaceUrb(ptr_size).get_cstruct()


class WDF_USB_INTERFACE_SELECT_SETTING_PARAMS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.Type = ct.c_uint32
        self.Types = InterfaceTypes


class Descriptor(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.ConfigurationDescriptor = Ptr
        self.InterfaceDescriptors = Ptr
        self.NumInterfaceDescriptors = ct.c_uint32


class Urb(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Urb = Ptr


class SingleInterface(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.NumberConfiguredPipes = ct.c_uint8
        self.ConfiguredUsbInterface = Ptr


class MultiInterface(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.NumberInterfaces = ct.c_uint8
        self.Pairs = Ptr
        self.NumberOfConfiguredInterfaces = ct.c_uint8


class Types(EmuUnion):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Descriptor = Descriptor(ptr_size).get_cstruct()
        self.Urb = Urb(ptr_size).get_cstruct()
        self.SingleInterface = SingleInterface(ptr_size).get_cstruct()
        self.MultiInterface = MultiInterface(ptr_size).get_cstruct()


class WDF_USB_DEVICE_SELECT_CONFIG_PARAMS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.Type = ct.c_uint32
        self.Types = Types


class WDF_IO_QUEUE_CONFIG(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Size = ct.c_uint32
        self.DispatchType = Ptr
        self.PowerManaged = Ptr
        self.AllowZeroLengthRequests = ct.c_uint8
        self.DefaultQueue = ct.c_uint8
        self.EvtIoDefault = Ptr
        self.EvtIoRead = Ptr
        self.EvtIoWrite = Ptr
        self.EvtIoDeviceControl = Ptr
        self.EvtIoInternalDeviceControl = Ptr
        self.EvtIoStop = Ptr
        self.EvtIoResume = Ptr
        self.EvtIoCanceledOnQueue = Ptr
        self.NumberOfPresentedRequests = ct.c_uint32
        self.Driver = Ptr


class WDFFUNCTIONS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.pfnWdfChildListCreate = Ptr
        self.pfnWdfChildListGetDevice = Ptr
        self.pfnWdfChildListRetrievePdo = Ptr
        self.pfnWdfChildListRetrieveAddressDescription = Ptr
        self.pfnWdfChildListBeginScan = Ptr
        self.pfnWdfChildListEndScan = Ptr
        self.pfnWdfChildListBeginIteration = Ptr
        self.pfnWdfChildListRetrieveNextDevice = Ptr
        self.pfnWdfChildListEndIteration = Ptr
        self.pfnWdfChildListAddOrUpdateChildDescriptionAsPresent = Ptr
        self.pfnWdfChildListUpdateChildDescriptionAsMissing = Ptr
        self.pfnWdfChildListUpdateAllChildDescriptionsAsPresent = Ptr
        self.pfnWdfChildListRequestChildEject = Ptr
        self.pfnWdfCollectionCreate = Ptr
        self.pfnWdfCollectionGetCount = Ptr
        self.pfnWdfCollectionAdd = Ptr
        self.pfnWdfCollectionRemove = Ptr
        self.pfnWdfCollectionRemoveItem = Ptr
        self.pfnWdfCollectionGetItem = Ptr
        self.pfnWdfCollectionGetFirstItem = Ptr
        self.pfnWdfCollectionGetLastItem = Ptr
        self.pfnWdfCommonBufferCreate = Ptr
        self.pfnWdfCommonBufferGetAlignedVirtualAddress = Ptr
        self.pfnWdfCommonBufferGetAlignedLogicalAddress = Ptr
        self.pfnWdfCommonBufferGetLength = Ptr
        self.pfnWdfControlDeviceInitAllocate = Ptr
        self.pfnWdfControlDeviceInitSetShutdownNotification = Ptr
        self.pfnWdfControlFinishInitializing = Ptr
        self.pfnWdfDeviceGetDeviceState = Ptr
        self.pfnWdfDeviceSetDeviceState = Ptr
        self.pfnWdfWdmDeviceGetWdfDeviceHandle = Ptr
        self.pfnWdfDeviceWdmGetDeviceObject = Ptr
        self.pfnWdfDeviceWdmGetAttachedDevice = Ptr
        self.pfnWdfDeviceWdmGetPhysicalDevice = Ptr
        self.pfnWdfDeviceWdmDispatchPreprocessedIrp = Ptr
        self.pfnWdfDeviceAddDependentUsageDeviceObject = Ptr
        self.pfnWdfDeviceAddRemovalRelationsPhysicalDevice = Ptr
        self.pfnWdfDeviceRemoveRemovalRelationsPhysicalDevice = Ptr
        self.pfnWdfDeviceClearRemovalRelationsDevices = Ptr
        self.pfnWdfDeviceGetDriver = Ptr
        self.pfnWdfDeviceRetrieveDeviceName = Ptr
        self.pfnWdfDeviceAssignMofResourceName = Ptr
        self.pfnWdfDeviceGetIoTarget = Ptr
        self.pfnWdfDeviceGetDevicePnpState = Ptr
        self.pfnWdfDeviceGetDevicePowerState = Ptr
        self.pfnWdfDeviceGetDevicePowerPolicyState = Ptr
        self.pfnWdfDeviceAssignS0IdleSettings = Ptr
        self.pfnWdfDeviceAssignSxWakeSettings = Ptr
        self.pfnWdfDeviceOpenRegistryKey = Ptr
        self.pfnWdfDeviceSetSpecialFileSupport = Ptr
        self.pfnWdfDeviceSetCharacteristics = Ptr
        self.pfnWdfDeviceGetCharacteristics = Ptr
        self.pfnWdfDeviceGetAlignmentRequirement = Ptr
        self.pfnWdfDeviceSetAlignmentRequirement = Ptr
        self.pfnWdfDeviceInitFree = Ptr
        self.pfnWdfDeviceInitSetPnpPowerEventCallbacks = Ptr
        self.pfnWdfDeviceInitSetPowerPolicyEventCallbacks = Ptr
        self.pfnWdfDeviceInitSetPowerPolicyOwnership = Ptr
        self.pfnWdfDeviceInitRegisterPnpStateChangeCallback = Ptr
        self.pfnWdfDeviceInitRegisterPowerStateChangeCallback = Ptr
        self.pfnWdfDeviceInitRegisterPowerPolicyStateChangeCallback = Ptr
        self.pfnWdfDeviceInitSetIoType = Ptr
        self.pfnWdfDeviceInitSetExclusive = Ptr
        self.pfnWdfDeviceInitSetPowerNotPageable = Ptr
        self.pfnWdfDeviceInitSetPowerPageable = Ptr
        self.pfnWdfDeviceInitSetPowerInrush = Ptr
        self.pfnWdfDeviceInitSetDeviceType = Ptr
        self.pfnWdfDeviceInitAssignName = Ptr
        self.pfnWdfDeviceInitAssignSDDLString = Ptr
        self.pfnWdfDeviceInitSetDeviceClass = Ptr
        self.pfnWdfDeviceInitSetCharacteristics = Ptr
        self.pfnWdfDeviceInitSetFileObjectConfig = Ptr
        self.pfnWdfDeviceInitSetRequestAttributes = Ptr
        self.pfnWdfDeviceInitAssignWdmIrpPreprocessCallback = Ptr
        self.pfnWdfDeviceInitSetIoInCallerContextCallback = Ptr
        self.pfnWdfDeviceCreate = Ptr
        self.pfnWdfDeviceSetStaticStopRemove = Ptr
        self.pfnWdfDeviceCreateDeviceInterface = Ptr
        self.pfnWdfDeviceSetDeviceInterfaceState = Ptr
        self.pfnWdfDeviceRetrieveDeviceInterfaceString = Ptr
        self.pfnWdfDeviceCreateSymbolicLink = Ptr
        self.pfnWdfDeviceQueryProperty = Ptr
        self.pfnWdfDeviceAllocAndQueryProperty = Ptr
        self.pfnWdfDeviceSetPnpCapabilities = Ptr
        self.pfnWdfDeviceSetPowerCapabilities = Ptr
        self.pfnWdfDeviceSetBusInformationForChildren = Ptr
        self.pfnWdfDeviceIndicateWakeStatus = Ptr
        self.pfnWdfDeviceSetFailed = Ptr
        self.pfnWdfDeviceStopIdleNoTrack = Ptr
        self.pfnWdfDeviceResumeIdleNoTrack = Ptr
        self.pfnWdfDeviceGetFileObject = Ptr
        self.pfnWdfDeviceEnqueueRequest = Ptr
        self.pfnWdfDeviceGetDefaultQueue = Ptr
        self.pfnWdfDeviceConfigureRequestDispatching = Ptr
        self.pfnWdfDmaEnablerCreate = Ptr
        self.pfnWdfDmaEnablerGetMaximumLength = Ptr
        self.pfnWdfDmaEnablerGetMaximumScatterGatherElements = Ptr
        self.pfnWdfDmaEnablerSetMaximumScatterGatherElements = Ptr
        self.pfnWdfDmaTransactionCreate = Ptr
        self.pfnWdfDmaTransactionInitialize = Ptr
        self.pfnWdfDmaTransactionInitializeUsingRequest = Ptr
        self.pfnWdfDmaTransactionExecute = Ptr
        self.pfnWdfDmaTransactionRelease = Ptr
        self.pfnWdfDmaTransactionDmaCompleted = Ptr
        self.pfnWdfDmaTransactionDmaCompletedWithLength = Ptr
        self.pfnWdfDmaTransactionDmaCompletedFinal = Ptr
        self.pfnWdfDmaTransactionGetBytesTransferred = Ptr
        self.pfnWdfDmaTransactionSetMaximumLength = Ptr
        self.pfnWdfDmaTransactionGetRequest = Ptr
        self.pfnWdfDmaTransactionGetCurrentDmaTransferLength = Ptr
        self.pfnWdfDmaTransactionGetDevice = Ptr
        self.pfnWdfDpcCreate = Ptr
        self.pfnWdfDpcEnqueue = Ptr
        self.pfnWdfDpcCancel = Ptr
        self.pfnWdfDpcGetParentObject = Ptr
        self.pfnWdfDpcWdmGetDpc = Ptr
        self.pfnWdfDriverCreate = Ptr
        self.pfnWdfDriverGetRegistryPath = Ptr
        self.pfnWdfDriverWdmGetDriverObject = Ptr
        self.pfnWdfDriverOpenParametersRegistryKey = Ptr
        self.pfnWdfWdmDriverGetWdfDriverHandle = Ptr
        self.pfnWdfDriverRegisterTraceInfo = Ptr
        self.pfnWdfDriverRetrieveVersionString = Ptr
        self.pfnWdfDriverIsVersionAvailable = Ptr
        self.pfnWdfFdoInitWdmGetPhysicalDevice = Ptr
        self.pfnWdfFdoInitOpenRegistryKey = Ptr
        self.pfnWdfFdoInitQueryProperty = Ptr
        self.pfnWdfFdoInitAllocAndQueryProperty = Ptr
        self.pfnWdfFdoInitSetEventCallbacks = Ptr
        self.pfnWdfFdoInitSetFilter = Ptr
        self.pfnWdfFdoInitSetDefaultChildListConfig = Ptr
        self.pfnWdfFdoQueryForInterface = Ptr
        self.pfnWdfFdoGetDefaultChildList = Ptr
        self.pfnWdfFdoAddStaticChild = Ptr
        self.pfnWdfFdoLockStaticChildListForIteration = Ptr
        self.pfnWdfFdoRetrieveNextStaticChild = Ptr
        self.pfnWdfFdoUnlockStaticChildListFromIteration = Ptr
        self.pfnWdfFileObjectGetFileName = Ptr
        self.pfnWdfFileObjectGetFlags = Ptr
        self.pfnWdfFileObjectGetDevice = Ptr
        self.pfnWdfFileObjectWdmGetFileObject = Ptr
        self.pfnWdfInterruptCreate = Ptr
        self.pfnWdfInterruptQueueDpcForIsr = Ptr
        self.pfnWdfInterruptSynchronize = Ptr
        self.pfnWdfInterruptAcquireLock = Ptr
        self.pfnWdfInterruptReleaseLock = Ptr
        self.pfnWdfInterruptEnable = Ptr
        self.pfnWdfInterruptDisable = Ptr
        self.pfnWdfInterruptWdmGetInterrupt = Ptr
        self.pfnWdfInterruptGetInfo = Ptr
        self.pfnWdfInterruptSetPolicy = Ptr
        self.pfnWdfInterruptGetDevice = Ptr
        self.pfnWdfIoQueueCreate = Ptr
        self.pfnWdfIoQueueGetState = Ptr
        self.pfnWdfIoQueueStart = Ptr
        self.pfnWdfIoQueueStop = Ptr
        self.pfnWdfIoQueueStopSynchronously = Ptr
        self.pfnWdfIoQueueGetDevice = Ptr
        self.pfnWdfIoQueueRetrieveNextRequest = Ptr
        self.pfnWdfIoQueueRetrieveRequestByFileObject = Ptr
        self.pfnWdfIoQueueFindRequest = Ptr
        self.pfnWdfIoQueueRetrieveFoundRequest = Ptr
        self.pfnWdfIoQueueDrainSynchronously = Ptr
        self.pfnWdfIoQueueDrain = Ptr
        self.pfnWdfIoQueuePurgeSynchronously = Ptr
        self.pfnWdfIoQueuePurge = Ptr
        self.pfnWdfIoQueueReadyNotify = Ptr
        self.pfnWdfIoTargetCreate = Ptr
        self.pfnWdfIoTargetOpen = Ptr
        self.pfnWdfIoTargetCloseForQueryRemove = Ptr
        self.pfnWdfIoTargetClose = Ptr
        self.pfnWdfIoTargetStart = Ptr
        self.pfnWdfIoTargetStop = Ptr
        self.pfnWdfIoTargetGetState = Ptr
        self.pfnWdfIoTargetGetDevice = Ptr
        self.pfnWdfIoTargetQueryTargetProperty = Ptr
        self.pfnWdfIoTargetAllocAndQueryTargetProperty = Ptr
        self.pfnWdfIoTargetQueryForInterface = Ptr
        self.pfnWdfIoTargetWdmGetTargetDeviceObject = Ptr
        self.pfnWdfIoTargetWdmGetTargetPhysicalDevice = Ptr
        self.pfnWdfIoTargetWdmGetTargetFileObject = Ptr
        self.pfnWdfIoTargetWdmGetTargetFileHandle = Ptr
        self.pfnWdfIoTargetSendReadSynchronously = Ptr
        self.pfnWdfIoTargetFormatRequestForRead = Ptr
        self.pfnWdfIoTargetSendWriteSynchronously = Ptr
        self.pfnWdfIoTargetFormatRequestForWrite = Ptr
        self.pfnWdfIoTargetSendIoctlSynchronously = Ptr
        self.pfnWdfIoTargetFormatRequestForIoctl = Ptr
        self.pfnWdfIoTargetSendInternalIoctlSynchronously = Ptr
        self.pfnWdfIoTargetFormatRequestForInternalIoctl = Ptr
        self.pfnWdfIoTargetSendInternalIoctlOthersSynchronously = Ptr
        self.pfnWdfIoTargetFormatRequestForInternalIoctlOthers = Ptr
        self.pfnWdfMemoryCreate = Ptr
        self.pfnWdfMemoryCreatePreallocated = Ptr
        self.pfnWdfMemoryGetBuffer = Ptr
        self.pfnWdfMemoryAssignBuffer = Ptr
        self.pfnWdfMemoryCopyToBuffer = Ptr
        self.pfnWdfMemoryCopyFromBuffer = Ptr
        self.pfnWdfLookasideListCreate = Ptr
        self.pfnWdfMemoryCreateFromLookaside = Ptr
        self.pfnWdfDeviceMiniportCreate = Ptr
        self.pfnWdfDriverMiniportUnload = Ptr
        self.pfnWdfObjectGetTypedContextWorker = Ptr
        self.pfnWdfObjectAllocateContext = Ptr
        self.pfnWdfObjectContextGetObject = Ptr
        self.pfnWdfObjectReferenceActual = Ptr
        self.pfnWdfObjectDereferenceActual = Ptr
        self.pfnWdfObjectCreate = Ptr
        self.pfnWdfObjectDelete = Ptr
        self.pfnWdfObjectQuery = Ptr
        self.pfnWdfPdoInitAllocate = Ptr
        self.pfnWdfPdoInitSetEventCallbacks = Ptr
        self.pfnWdfPdoInitAssignDeviceID = Ptr
        self.pfnWdfPdoInitAssignInstanceID = Ptr
        self.pfnWdfPdoInitAddHardwareID = Ptr
        self.pfnWdfPdoInitAddCompatibleID = Ptr
        self.pfnWdfPdoInitAddDeviceText = Ptr
        self.pfnWdfPdoInitSetDefaultLocale = Ptr
        self.pfnWdfPdoInitAssignRawDevice = Ptr
        self.pfnWdfPdoMarkMissing = Ptr
        self.pfnWdfPdoRequestEject = Ptr
        self.pfnWdfPdoGetParent = Ptr
        self.pfnWdfPdoRetrieveIdentificationDescription = Ptr
        self.pfnWdfPdoRetrieveAddressDescription = Ptr
        self.pfnWdfPdoUpdateAddressDescription = Ptr
        self.pfnWdfPdoAddEjectionRelationsPhysicalDevice = Ptr
        self.pfnWdfPdoRemoveEjectionRelationsPhysicalDevice = Ptr
        self.pfnWdfPdoClearEjectionRelationsDevices = Ptr
        self.pfnWdfDeviceAddQueryInterface = Ptr
        self.pfnWdfRegistryOpenKey = Ptr
        self.pfnWdfRegistryCreateKey = Ptr
        self.pfnWdfRegistryClose = Ptr
        self.pfnWdfRegistryWdmGetHandle = Ptr
        self.pfnWdfRegistryRemoveKey = Ptr
        self.pfnWdfRegistryRemoveValue = Ptr
        self.pfnWdfRegistryQueryValue = Ptr
        self.pfnWdfRegistryQueryMemory = Ptr
        self.pfnWdfRegistryQueryMultiString = Ptr
        self.pfnWdfRegistryQueryUnicodeString = Ptr
        self.pfnWdfRegistryQueryString = Ptr
        self.pfnWdfRegistryQueryULong = Ptr
        self.pfnWdfRegistryAssignValue = Ptr
        self.pfnWdfRegistryAssignMemory = Ptr
        self.pfnWdfRegistryAssignMultiString = Ptr
        self.pfnWdfRegistryAssignUnicodeString = Ptr
        self.pfnWdfRegistryAssignString = Ptr
        self.pfnWdfRegistryAssignULong = Ptr
        self.pfnWdfRequestCreate = Ptr
        self.pfnWdfRequestCreateFromIrp = Ptr
        self.pfnWdfRequestReuse = Ptr
        self.pfnWdfRequestChangeTarget = Ptr
        self.pfnWdfRequestFormatRequestUsingCurrentType = Ptr
        self.pfnWdfRequestWdmFormatUsingStackLocation = Ptr
        self.pfnWdfRequestSend = Ptr
        self.pfnWdfRequestGetStatus = Ptr
        self.pfnWdfRequestMarkCancelable = Ptr
        self.pfnWdfRequestUnmarkCancelable = Ptr
        self.pfnWdfRequestIsCanceled = Ptr
        self.pfnWdfRequestCancelSentRequest = Ptr
        self.pfnWdfRequestIsFrom32BitProcess = Ptr
        self.pfnWdfRequestSetCompletionRoutine = Ptr
        self.pfnWdfRequestGetCompletionParams = Ptr
        self.pfnWdfRequestAllocateTimer = Ptr
        self.pfnWdfRequestComplete = Ptr
        self.pfnWdfRequestCompleteWithPriorityBoost = Ptr
        self.pfnWdfRequestCompleteWithInformation = Ptr
        self.pfnWdfRequestGetParameters = Ptr
        self.pfnWdfRequestRetrieveInputMemory = Ptr
        self.pfnWdfRequestRetrieveOutputMemory = Ptr
        self.pfnWdfRequestRetrieveInputBuffer = Ptr
        self.pfnWdfRequestRetrieveOutputBuffer = Ptr
        self.pfnWdfRequestRetrieveInputWdmMdl = Ptr
        self.pfnWdfRequestRetrieveOutputWdmMdl = Ptr
        self.pfnWdfRequestRetrieveUnsafeUserInputBuffer = Ptr
        self.pfnWdfRequestRetrieveUnsafeUserOutputBuffer = Ptr
        self.pfnWdfRequestSetInformation = Ptr
        self.pfnWdfRequestGetInformation = Ptr
        self.pfnWdfRequestGetFileObject = Ptr
        self.pfnWdfRequestProbeAndLockUserBufferForRead = Ptr
        self.pfnWdfRequestProbeAndLockUserBufferForWrite = Ptr
        self.pfnWdfRequestGetRequestorMode = Ptr
        self.pfnWdfRequestForwardToIoQueue = Ptr
        self.pfnWdfRequestGetIoQueue = Ptr
        self.pfnWdfRequestRequeue = Ptr
        self.pfnWdfRequestStopAcknowledge = Ptr
        self.pfnWdfRequestWdmGetIrp = Ptr
        self.pfnWdfIoResourceRequirementsListSetSlotNumber = Ptr
        self.pfnWdfIoResourceRequirementsListSetInterfaceType = Ptr
        self.pfnWdfIoResourceRequirementsListAppendIoResList = Ptr
        self.pfnWdfIoResourceRequirementsListInsertIoResList = Ptr
        self.pfnWdfIoResourceRequirementsListGetCount = Ptr
        self.pfnWdfIoResourceRequirementsListGetIoResList = Ptr
        self.pfnWdfIoResourceRequirementsListRemove = Ptr
        self.pfnWdfIoResourceRequirementsListRemoveByIoResList = Ptr
        self.pfnWdfIoResourceListCreate = Ptr
        self.pfnWdfIoResourceListAppendDescriptor = Ptr
        self.pfnWdfIoResourceListInsertDescriptor = Ptr
        self.pfnWdfIoResourceListUpdateDescriptor = Ptr
        self.pfnWdfIoResourceListGetCount = Ptr
        self.pfnWdfIoResourceListGetDescriptor = Ptr
        self.pfnWdfIoResourceListRemove = Ptr
        self.pfnWdfIoResourceListRemoveByDescriptor = Ptr
        self.pfnWdfCmResourceListAppendDescriptor = Ptr
        self.pfnWdfCmResourceListInsertDescriptor = Ptr
        self.pfnWdfCmResourceListGetCount = Ptr
        self.pfnWdfCmResourceListGetDescriptor = Ptr
        self.pfnWdfCmResourceListRemove = Ptr
        self.pfnWdfCmResourceListRemoveByDescriptor = Ptr
        self.pfnWdfStringCreate = Ptr
        self.pfnWdfStringGetUnicodeString = Ptr
        self.pfnWdfObjectAcquireLock = Ptr
        self.pfnWdfObjectReleaseLock = Ptr
        self.pfnWdfWaitLockCreate = Ptr
        self.pfnWdfWaitLockAcquire = Ptr
        self.pfnWdfWaitLockRelease = Ptr
        self.pfnWdfSpinLockCreate = Ptr
        self.pfnWdfSpinLockAcquire = Ptr
        self.pfnWdfSpinLockRelease = Ptr
        self.pfnWdfTimerCreate = Ptr
        self.pfnWdfTimerStart = Ptr
        self.pfnWdfTimerStop = Ptr
        self.pfnWdfTimerGetParentObject = Ptr
        self.pfnWdfUsbTargetDeviceCreate = Ptr
        self.pfnWdfUsbTargetDeviceRetrieveInformation = Ptr
        self.pfnWdfUsbTargetDeviceGetDeviceDescriptor = Ptr
        self.pfnWdfUsbTargetDeviceRetrieveConfigDescriptor = Ptr
        self.pfnWdfUsbTargetDeviceQueryString = Ptr
        self.pfnWdfUsbTargetDeviceAllocAndQueryString = Ptr
        self.pfnWdfUsbTargetDeviceFormatRequestForString = Ptr
        self.pfnWdfUsbTargetDeviceGetNumInterfaces = Ptr
        self.pfnWdfUsbTargetDeviceSelectConfig = Ptr
        self.pfnWdfUsbTargetDeviceWdmGetConfigurationHandle = Ptr
        self.pfnWdfUsbTargetDeviceRetrieveCurrentFrameNumber = Ptr
        self.pfnWdfUsbTargetDeviceSendControlTransferSynchronously = Ptr
        self.pfnWdfUsbTargetDeviceFormatRequestForControlTransfer = Ptr
        self.pfnWdfUsbTargetDeviceIsConnectedSynchronous = Ptr
        self.pfnWdfUsbTargetDeviceResetPortSynchronously = Ptr
        self.pfnWdfUsbTargetDeviceCyclePortSynchronously = Ptr
        self.pfnWdfUsbTargetDeviceFormatRequestForCyclePort = Ptr
        self.pfnWdfUsbTargetDeviceSendUrbSynchronously = Ptr
        self.pfnWdfUsbTargetDeviceFormatRequestForUrb = Ptr
        self.pfnWdfUsbTargetPipeGetInformation = Ptr
        self.pfnWdfUsbTargetPipeIsInEndpoint = Ptr
        self.pfnWdfUsbTargetPipeIsOutEndpoint = Ptr
        self.pfnWdfUsbTargetPipeGetType = Ptr
        self.pfnWdfUsbTargetPipeSetNoMaximumPacketSizeCheck = Ptr
        self.pfnWdfUsbTargetPipeWriteSynchronously = Ptr
        self.pfnWdfUsbTargetPipeFormatRequestForWrite = Ptr
        self.pfnWdfUsbTargetPipeReadSynchronously = Ptr
        self.pfnWdfUsbTargetPipeFormatRequestForRead = Ptr
        self.pfnWdfUsbTargetPipeConfigContinuousReader = Ptr
        self.pfnWdfUsbTargetPipeAbortSynchronously = Ptr
        self.pfnWdfUsbTargetPipeFormatRequestForAbort = Ptr
        self.pfnWdfUsbTargetPipeResetSynchronously = Ptr
        self.pfnWdfUsbTargetPipeFormatRequestForReset = Ptr
        self.pfnWdfUsbTargetPipeSendUrbSynchronously = Ptr
        self.pfnWdfUsbTargetPipeFormatRequestForUrb = Ptr
        self.pfnWdfUsbInterfaceGetInterfaceNumber = Ptr
        self.pfnWdfUsbInterfaceGetNumEndpoints = Ptr
        self.pfnWdfUsbInterfaceGetDescriptor = Ptr
        self.pfnWdfUsbInterfaceSelectSetting = Ptr
        self.pfnWdfUsbInterfaceGetEndpointInformation = Ptr
        self.pfnWdfUsbTargetDeviceGetInterface = Ptr
        self.pfnWdfUsbInterfaceGetConfiguredSettingIndex = Ptr
        self.pfnWdfUsbInterfaceGetNumConfiguredPipes = Ptr
        self.pfnWdfUsbInterfaceGetConfiguredPipe = Ptr
        self.pfnWdfUsbTargetPipeWdmGetPipeHandle = Ptr
        self.pfnWdfVerifierDbgBreakPoint = Ptr
        self.pfnWdfVerifierKeBugCheck = Ptr
        self.pfnWdfWmiProviderCreate = Ptr
        self.pfnWdfWmiProviderGetDevice = Ptr
        self.pfnWdfWmiProviderIsEnabled = Ptr
        self.pfnWdfWmiProviderGetTracingHandle = Ptr
        self.unknown0 = Ptr * 13
        self.pfnWdfUsbInterfaceGetNumSettings = Ptr
        self.unknown1 = Ptr * 34
        self.pfnWdfUsbTargetDeviceCreateWithParameters = Ptr



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Enum
import ctypes as ct

REG_NONE = 0
REG_SZ = 1
REG_EXPAND_SZ = 2
REG_BINARY = 3
REG_DWORD = 4
REG_MULTI_SZ = 7
REG_QWORD = 11

RTL_REGISTRY_ABSOLUTE = 0
RTL_REGISTRY_SERVICES = 1
RTL_REGISTRY_CONTROL = 2
RTL_REGISTRY_WINDOWS_NT = 3
RTL_REGISTRY_DEVICEMAP = 4
RTL_REGISTRY_USER = 5
RTL_REGISTRY_MAXIMUM = 6

HKEY_CLASSES_ROOT = 0x80000000
HKEY_CURRENT_USER = 0x80000001
HKEY_LOCAL_MACHINE = 0x80000002
HKEY_USERS = 0x80000003


KEY_VALUE_INFORMATION_CLASS = Enum()
KEY_VALUE_INFORMATION_CLASS.KeyValueBasicInformation = 0x00
KEY_VALUE_INFORMATION_CLASS.KeyValueFullInformation = 0x01
KEY_VALUE_INFORMATION_CLASS.KeyValuePartialInformation = 0x02
KEY_VALUE_INFORMATION_CLASS.KeyValueFullInformationAlign64 = 0x03
KEY_VALUE_INFORMATION_CLASS.KeyValuePartialInformationAlign64 = 0x04
KEY_VALUE_INFORMATION_CLASS.KeyValueLayerInformation = 0x05
KEY_VALUE_INFORMATION_CLASS.MaxKeyValueInfoClass = 0x06


class KEY_VALUE_PARTIAL_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.TitleIndex = ct.c_uint32
        self.Type = ct.c_uint32
        self.DataLength = ct.c_uint32
        # Data[1]


class KEY_VALUE_BASIC_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.TitleIndex = ct.c_uint32
        self.Type = ct.c_uint32
        self.NameLength = ct.c_uint32
        # Name[1]


class KEY_VALUE_FULL_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.TitleIndex = ct.c_uint32
        self.Type = ct.c_uint32
        self.DataOffset = ct.c_uint32
        self.DataLength = ct.c_uint32
        self.NameLength = ct.c_uint32


def get_defines(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k


def get_flag_value(flag):
    return globals().get(flag)


def get_value_type(define):
    return get_defines(define, prefix='REG_')


def get_hkey_type(define):
    return get_defines(define, prefix='HKEY_')



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

with open("speakeasy/version.py", encoding="utf8") as fp:
    vl = fp.readline()
    gv, ver_num = vl.split("=")
    if gv.strip() != "__version__":
        raise Exception("Invalid version file found")
    version = ver_num.strip().strip("\"'")

with open("requirements.txt", encoding="utf-8") as f:
    requirements = [line.strip() for line in f.readlines()]

setuptools.setup(
    name="speakeasy-emulator",
    author="Andrew Davis",
    description="Speakeasy malware emulation framework",
    version=version,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    url="https://github.com/fireeye/speakeasy",
    include_package_data=True,
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["speakeasy=speakeasy.cli:main"]},
)



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import base64
import fnmatch
import speakeasy.winenv.defs.registry.reg as regdefs
from speakeasy.errors import RegistryEmuError

HKEY_CLASSES_ROOT = 0x80000000
HKEY_CURRENT_USER = 0x80000001
HKEY_LOCAL_MACHINE = 0x80000002
HKEY_USERS = 0x80000003


class RegValue(object):
    """
    Represents a registry value
    """
    def __init__(self, name, val_type, data):
        self.name = name
        self.type = val_type
        self.data = self.normalize_value(val_type, data)

    def normalize_value(self, val_type, data):
        """
        Convert registry values to python types
        """
        if val_type in (regdefs.REG_EXPAND_SZ, regdefs.REG_MULTI_SZ, regdefs.REG_SZ):
            if not isinstance(data, str):
                raise RegistryEmuError('Invalid registry value expected string')
            return data
        elif val_type in (regdefs.REG_DWORD, regdefs.REG_QWORD):
            if isinstance(data, str):
                return int(data, 0)
            elif isinstance(data, int):
                return data
        elif val_type == regdefs.REG_BINARY:
            # Binary data is expected to be base64'd
            return base64.b64encode(data.encode('utf-8'))
        else:
            return data

    def get_name(self):
        return self.name

    def get_type(self):
        return self.type

    def get_data(self):
        return self.data


class RegKey(object):
    """
    Represents a registry key
    """
    curr_handle = 0x180

    def __init__(self, path):
        self.path = path
        self.values = []

    def get_handle(self):
        hkey = RegKey.curr_handle
        RegKey.curr_handle += 4
        return hkey

    def get_path(self):
        return self.path

    def create_value(self, name, val_type, value):
        val = RegValue(name, val_type, value)
        self.values.append(val)
        return val

    def get_values(self):
        return self.values

    def get_value(self, val_name):
        if not val_name:
            val_name = 'default'
        for v in self.values:
            if val_name.lower() == v.get_name().lower():
                return v


class RegistryManager(object):
    """
    Manages the emulation of the windows registry. This includes creating keys, subkeys and values
    """
    def __init__(self, config=None):
        super(RegistryManager, self).__init__()
        self.reg_handles = {}
        self.keys = []
        self.config = config
        self.reg_tree = []

        for hk in (HKEY_CLASSES_ROOT, HKEY_CURRENT_USER,
                   HKEY_LOCAL_MACHINE, HKEY_USERS):
            path = regdefs.get_hkey_type(hk)
            key = self.create_key(path)
            self.reg_handles.update({hk: key})

    def normalize_reg_path(self, path):
        new = path
        if path:
            roots = ('\\registry\\machine\\', 'hklm\\')
            for r in roots:
                if path.lower().startswith(r):
                    new = 'HKEY_LOCAL_MACHINE\\' + path[len(r):]
                    return new
        return path

    def get_key_from_handle(self, handle):
        return self.reg_handles.get(handle)

    def get_key_from_path(self, path):
        path = self.normalize_reg_path(path)
        for key in self.keys:
            if fnmatch.fnmatch(key.get_path().lower(), path.lower()):
                return key
        return None

    def is_key_a_parent_key(self, path):
        for key in self.keys:
            if key.get_path().lower().startswith(path.lower()):
                return True
        return False

    def get_subkeys(self, key):
        # TODO: once we revamp the registry emulation,
        # make this better

        parent_path = key.get_path()
        subkeys = []
        for k in self.keys:
            test_path = k.get_path()
            if test_path.lower().startswith(parent_path.lower()):
                sub = test_path[len(parent_path):]
                if sub.startswith('\\'):
                    sub = sub[1:]

                end_slash = sub.find('\\')
                if end_slash >= 0:
                    sub = sub[:end_slash]

                if not sub:
                    continue

                subkeys.append(sub)

        return subkeys

    def get_key_from_config(self, path):
        """
        See if the emulator config file contains a handler for the requested registry path
        """
        for key in self.config.get('keys', []):
            if key['path'].lower() == path.lower():
                new_key = RegKey(path)
                for value in key.get('values', []):
                    val_type = value.get('type')
                    vts = regdefs.get_flag_value(val_type)  # noqa

                    val_name = value.get('name', '')
                    data = value.get('data')
                    new_key.create_value(val_name, val_type, data)
                return new_key
        return None

    def create_key(self, path):
        """
        Create a registry key
        """

        path = self.normalize_reg_path(path)
        # Does this key already exist?
        key = self.get_key_from_path(path)
        if key:
            return key

        # Does this key exist in our config
        key = self.get_key_from_config(path)
        if key:
            return key

        key = RegKey(path)
        self.keys.append(key)
        return key

    def open_key(self, path, create=False):
        """
        Open or optionally create a registry key
        """
        hnd = None
        path = self.normalize_reg_path(path)
        # Does the key already exist?
        key = self.get_key_from_path(path)
        if key:
            hnd = key.get_handle()
            self.reg_handles.update({hnd: key})
            return hnd

        # Does this key exist in our config
        key = self.get_key_from_config(path)
        if key:
            hnd = key.get_handle()
            self.reg_handles.update({hnd: key})
            return hnd

        # If we are instructed to create the key, do so
        if create or self.is_key_a_parent_key(path):
            key = RegKey(path)
            hnd = key.get_handle()
            self.reg_handles.update({hnd: key})
            self.keys.append(key)
        return hnd



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.windows.com as com
import speakeasy.winenv.defs.windows.windows as windefs

from .. import api


class Ole32(api.ApiHandler):

    name = 'ole32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Ole32, self).__init__(emu)
        self.funcs = {}
        self.data = {}
        super(Ole32, self).__get_hook_attrs__(self)
        self.netman = emu.get_network_manager()
        self.names = {}

    @apihook('OleInitialize', argc=1)
    def OleInitialize(self, emu, argv, ctx={}):
        """
        HRESULT OleInitialize(
            IN LPVOID pvReserved
        );
        """

        rv = windefs.S_OK

        return rv

    @apihook('CoInitialize', argc=1)
    def CoInitialize(self, emu, argv, ctx={}):
        """
        HRESULT CoInitialize(
          LPVOID pvReserved
        );
        """

        rv = windefs.S_OK

        return rv

    @apihook('CoInitializeEx', argc=2)
    def CoInitializeEx(self, emu, argv, ctx={}):
        """
        HRESULT CoInitializeEx(
          LPVOID pvReserved,
          DWORD  dwCoInit
        );
        """

        rv = windefs.S_OK

        return rv

    @apihook('CoUninitialize', argc=0)
    def CoUninitialize(self, emu, argv, ctx={}):
        """
        void CoUninitialize();
        """

    @apihook('CoInitializeSecurity', argc=9)
    def CoInitializeSecurity(self, emu, argv, ctx={}):
        """
        HRESULT CoInitializeSecurity(
          PSECURITY_DESCRIPTOR        pSecDesc,
          LONG                        cAuthSvc,
          SOLE_AUTHENTICATION_SERVICE *asAuthSvc,
          void                        *pReserved1,
          DWORD                       dwAuthnLevel,
          DWORD                       dwImpLevel,
          void                        *pAuthList,
          DWORD                       dwCapabilities,
          void                        *pReserved3
        );
        """

        rv = windefs.S_OK

        authn_level = com.get_define_int(argv[4])
        if authn_level:
            argv[4] = authn_level

        imp_level = com.get_define_int(argv[5])
        if imp_level:
            argv[5] = imp_level

        return rv

    @apihook('CoCreateInstance', argc=5)
    def CoCreateInstance(self, emu, argv, ctx={}):
        """
        HRESULT CoCreateInstance(
          REFCLSID  rclsid,
          LPUNKNOWN pUnkOuter,
          DWORD     dwClsContext,
          REFIID    riid,
          LPVOID    *ppv
        );
        """
        rclsid, pUnkOuter, dwClsContext, riid, ppv = argv
        rv = windefs.S_OK

        clsid_bytes = self.mem_read(rclsid, self.sizeof(windefs.GUID()))
        clsid_str = com.convert_guid_bytes_to_str(clsid_bytes)
        clsid_name = com.get_clsid(clsid_str)
        if clsid_name:
            argv[0] = clsid_name
            riid_bytes = self.mem_read(riid, self.sizeof(windefs.GUID()))
            riid_str = com.convert_guid_bytes_to_str(riid_bytes)
            iid_name = com.get_iid(riid_str)
            if iid_name:
                argv[3] = iid_name
                if ppv:
                    ci = emu.com.get_interface(emu, emu.get_ptr_size(), iid_name.replace('IID_', ''))
                    pv = self.mem_alloc(emu.get_ptr_size(), tag='emu.COM.pv_%s' % iid_name)
                    self.mem_write(pv, ci.address.to_bytes(emu.get_ptr_size(), 'little'))
                    self.mem_write(ppv, pv.to_bytes(emu.get_ptr_size(), 'little'))
            else:
                self.emu.logger.info('Unsupported COM IID %s', riid)
        else:
            self.emu.logger.info('Unsupported COM CLSID %s', clsid_str)

        return rv

    @apihook('CoSetProxyBlanket', argc=8)
    def CoSetProxyBlanket(self, emu, argv, ctx={}):
        """
        HRESULT CoSetProxyBlanket(
            IUnknown                 *pProxy,
            DWORD                    dwAuthnSvc,
            DWORD                    dwAuthzSvc,
            OLECHAR                  *pServerPrincName,
            DWORD                    dwAuthnLevel,
            DWORD                    dwImpLevel,
            RPC_AUTH_IDENTITY_HANDLE pAuthInfo,
            DWORD                    dwCapabilities
        );
        """
        return 1

    @apihook('StringFromCLSID', argc=2)
    def StringFromCLSID(self, emu, argv, ctx={}):
        """
        HRESULT StringFromCLSID(
        REFCLSID rclsid,
        LPOLESTR *lplpsz
        );
        """

        rclsid, lplpsz = argv
        rv = windefs.S_OK

        guid = self.mem_read(rclsid, self.sizeof(windefs.GUID()))
        u = com.convert_guid_bytes_to_str(guid)
        argv[1] = u
        u = (u + '\x00').encode('utf-16le')

        ptr = self.mem_alloc(len(u), tag='api.StringFromCLSID')

        if lplpsz:
            self.mem_write(lplpsz, ptr.to_bytes(emu.get_ptr_size(), 'little'))

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr
import ctypes as ct

NTE_BAD_ALGID = 0x80090008

SERVICE_WIN32 = 0x30

SERVICE_ACTIVE = 0x1


class SERVICE_TABLE_ENTRY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.lpServiceName = Ptr
        self.lpServiceProc = Ptr

class HCRYPTKEY(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Algid = ct.c_uint32
        self.keylen = ct.c_uint32
        self.keyp = Ptr


def get_define_int(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k
